/* ============================= C MeatAxe ==================================
   File:        $Id: bsop.c,v 1.1.1.1 2000/01/07 16:42:26 mringe Exp $
   Comment:     Bit string operations.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <string.h>
   
/* --------------------------------------------------------------------------
   Local data
   -------------------------------------------------------------------------- */

MTX_DEFINE_FILE_INFO

#define BPL (sizeof(long) * 8)		/* Number of bits in a long */



/*{{{ --------------------------------------------------------------------------
!section obj.bs
!function BsClearAll "Clear a bit string"
!arg bs
    Pointer to the bit string.
!return
    $0$ on success, $-1$ on error.
!description
    This function clears all bits in a bit string.
!seealso
-------------------------------------------------------------------------- }}}*/

int BsClearAll(BitString_t *bs)

{
    if (!BsIsValid(bs))
	return -1;
    memset(bs->Data,0,bs->BufSize * sizeof(long));
    return 0;
}




/*{{{ --------------------------------------------------------------------------
!function BsCompare "Compare two bit strings"
!arg a
    Pointer to the first bit string.
!arg b
    Pointer to the second bit string.
!return
    $0$ if the bit strings are equal. Otherwise the return value is different 
    from zero.
!description
    This function compares two bit strings and returns zero if they are equal.
!seealso
-------------------------------------------------------------------------- }}}*/

int BsCompare(const BitString_t *a, const BitString_t *b)

{
    int i;
    if (!BsIsValid(a) || !BsIsValid(b))
	return -1;
    i = a->Size - b->Size;
    if (i != 0)
	return i;
    return memcmp(a->Data,b->Data,a->BufSize * sizeof(long));
}




/*{{{ --------------------------------------------------------------------------
!function BsCopy "Copy a bit string"
!synopsis 
    BitString_t *BsCopy(BitString_t *dest, const BitString_t *src);
!arg dest
    Pointer to the destination bit string.
!arg src
    Pointer to the source bit string.
!return
    |dest| on success, |NULL| on error.
!description
    This function copy a bit string from |src| to |dest|. The bit strings must
    have the same size.
!seealso BsDup
-------------------------------------------------------------------------- }}}*/

BitString_t *BsCopy(BitString_t *dest, const BitString_t *src)

{	
    if (!BsIsValid(dest) || !BsIsValid(src))
	return NULL;
    if (dest->Size != src->Size)
    {
	MTX_ERROR1("%E",MTX_ERR_INCOMPAT);
	return NULL;
    }
    memcpy(dest->Data,src->Data,src->BufSize * sizeof(long));
    return dest;
}



/*{{{ --------------------------------------------------------------------------
!function BsSet "Set one bit in a bit string"
!synopsis
    int BsSet(BitString_t *bs, int i)
!arg bs
    Pointer to the bit string.
!arg i
    Index of bit to set (starting from 0).
!return
    $0$ on sucess, $-1$ on error.
!description
    This function sets a single bit in a bit string. |i| is the index of
    the bit to set, the first bit having index $0$. |i| must be less than
    the bit string's size, that was specified when creating the bit string.
    It is not possible to change the size of an existing bit string.
!seealso BsClear BsTest
-------------------------------------------------------------------------- }}}*/

#ifdef DEBUG

int BsSet(BitString_t *bs, int i)

{
    if (!BsIsValid(bs))
	return -1;
    if (i < 0 || i >= bs->Size)
    {
	MTX_ERROR2("i=%d: %E",i,MTX_ERR_BADARG);
	return -1;
    }
    bs->Data[i / BPL] |= 1L << (i % BPL);
    return 0;
}

#endif


/*{{{ --------------------------------------------------------------------------
!function BsClear "Clear one bit in a bit string"
!synopsis
    int BsClear(BitString_t *bs, int i)
!arg bs
    Pointer to the bit string.
!arg i
    Index of bit to clear (starting from 0).
!return
    $0$ on success, $-1$ on error.
!description
    This function clears a single bit in a bit string.
!seealso @bsbit
-------------------------------------------------------------------------- }}}*/

#ifdef DEBUG

int BsClear(BitString_t *bs, int i)

{	
    if (!BsIsValid(bs))
	return -1;
    if (i < 0 || i >= bs->Size)
    {
	MTX_ERROR2("i=%d: %E",i,MTX_ERR_BADARG);
	return -1;
    }
    bs->Data[i / BPL] &= ~(1L << (i % BPL));
    return 0;
}

#endif


/*{{{ --------------------------------------------------------------------------
!function BsTest "Test a bit in a bit string"
!synopsis
    int BsTest(const BitString_t *bs, int i)
!arg bs
    Pointer to the bit string.
!arg i
    Index of bit to test (starting from 0).
!return
    $1$ if the bit is set, $0$ if the bit is not set, $-1$ on error.
!description
    This function tests a single bit in a bit string. The return value is
    $1$ id the bit is set and $0$ otherwise.
!seealso @bsbit
-------------------------------------------------------------------------- }}}*/

#ifdef DEBUG

int BsTest(const BitString_t *bs, int i)

{	
    if (!BsIsValid(bs))
	return -1;
    if (i < 0 || i >= bs->Size)
    {
	MTX_ERROR2("i=%d: %E",i,MTX_ERR_BADARG);
	return -1;
    }
    return (bs->Data[i / BPL] & (1L << (i % BPL))) != 0 ? 1 : 0;
}


#endif
