/* ============================= C MeatAxe ==================================
   File:        $Id: ffio.c,v 1.2 2001/04/04 17:59:38 mringe Exp $
   Comment:     I/o for vectors and matrices.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include <string.h>
#include <stdlib.h>
#include "meataxe.h"


MTX_DEFINE_FILE_INFO


/*{{{ --------------------------------------------------------------------------
!section kernel.ff.io
!function FfReadRows "Read rows"
!synopsis 
    int FfReadRows(FILE *f, PTR buf, int n)
!arg f
    Pointer to File.
!arg buf
    Pointer to data buffer.
!arg n
    Number of rows to read.
!return
    Number of rows that were actually read, or $-1$ on error.
!description
    This function reads 1 or more rows from a file.
    The row size must have been set before.
!seealso FfSeekRow FfWriteRows
-------------------------------------------------------------------------- }}}*/

int FfReadRows(FILE *f, PTR buf, int n)

{
    int i;
    register char *b = (char *) buf;

    /* Handle special case <FfNoc> = 0
       ------------------------------- */
    if (FfNoc == 0)
	return n;

    /* Read rows one by one
       -------------------- */
    for (i = 0; i < n; ++i)
    {
        if (fread(b,FfTrueRowSize(FfNoc),1,f) != 1) break;
	b += FfCurrentRowSize;
    }
    if (ferror(f)) 
	MTX_ERROR("Read failed: %S");
    return i;
}



/*{{{ --------------------------------------------------------------------------
!function FfWriteRows "Write rows"
!synopsis 
    int FfWriteRows(FILE *f, PTR  buf, int n);
!arg f
    Pointer to File.
!arg buf
    Pointer to data buffer.
!arg n
    Number of rows to write.
!return
    The function returns the number of rows that were successully written.
    A return value different from |n| indicates an error.
!description
    This function writes 1 or more rows to a file.
    The row size must have been set before.
!seealso FfSeekRow FfReadRows
-------------------------------------------------------------------------- }}}*/

int FfWriteRows(FILE *f, PTR  buf, int n)

{
    int i;
    register char *b = (char *) buf;

    /* Handle special case <FfNoc> = 0
       ------------------------------- */
    if (FfNoc == 0)
	return n;

    /* Write rows one by one
       --------------------- */
    for (i = 0; i < n; ++i)
    {
        if (fwrite(b,FfTrueRowSize(FfNoc),1,f) != 1) break;
	b += FfCurrentRowSize;
    }
    if (ferror(f)) 
	MTX_ERROR("Write failed: %S");
    return i;
}




/*{{{ --------------------------------------------------------------------------
!function FfSeekRow "Move to a row"
!arg f
    Pointer to File.
!arg pos
    Row number (0-based).
!return
    The function returns $0$ on success and $-1$ on error.
!description
    This function sets the read/write pointer of file \verb"f" to position 
    \verb"pos". I.e., the next |FfReadRows()| or |FfWriteRows()| will start
    at the specified row.
    Note that row numbers start with 0. If |pos| is different 
    from 0, the row size must have been set before with |zsetlen()|.
    
    You should always use |FfSeekRow()| rather than |fseek()| because
    |FfSeekRow()| knows about {\MeatAxe} file headers and adjusts the
    file pointer appropriately. 
!seealso FfReadRows FfWriteRows
-------------------------------------------------------------------------- }}}*/

int FfSeekRow(FILE *f, int pos)

{
    int addr;

    if (FfOrder != -1)
	addr = FfTrueRowSize(FfNoc) * pos + 12;
    else
	addr = FfNoc * 4 * pos + 12;
    if (SysFseek(f,addr) == -1)
    {
	MTX_ERROR("Seek failed: %S");
	return -1;
    }
    return 0;
}


/*{{{ --------------------------------------------------------------------------
!function FfReadHeader "Open file and read header"
!arg name
    File name.
!arg field
    Pointer to buffer for first header entry (usually the field order).
!arg nor
    Pointer to buffer for second header entry (usually the number of rows).
!arg noc
    Pointer to buffer for third header entry (usually the number of columns).
!return
    Pointer to open file, or |NULL| on error.
!description
    This function opens a data file for input and reads the file header
    (3 integers). The exact meaning of the header values
    depends on the file type. For a matrix they are `field order', `number 
    of rows' and `number of columns'.
!seealso FfWriteHeader
-------------------------------------------------------------------------- }}}*/

FILE *FfReadHeader(const char *name, int *field, int *nor, int *noc)

{
    FILE *fd;
    long header[3];

    /* Open the file
       ------------- */
    fd = SysFopen(name,FM_READ);
    if (fd == NULL)
	return NULL;

    /* Read the file header
       -------------------- */
    if (SysReadLong(fd,header,3) != 3)
    {
	fclose(fd);
	MTX_ERROR1("%s: Error reading file header",name);
	return NULL;
    }

    /* Check header
       ------------ */
    if (header[0] > 256 || header[1] < 0 || header[2] < 0)
    {	
	MTX_ERROR1("%s: Invalid header, possibly non-MeatAxe file",name);
	fclose(fd);
	return NULL;
    }

    /* Store header values in user-supplied buffers
       -------------------------------------------- */
    *field = (int)header[0];
    *nor = (int)header[1];
    *noc = (int)header[2];

    return fd;
}



/*{{{ --------------------------------------------------------------------------
!function FfWriteHeader "Open file and write header"
!synopsis 
    FILE *FfWriteHeader(const char *name, int field, int nrows, int ncols);
!arg name
    File name.
!arg field
    First header entry (usually the field order).
!arg nor
    Second header entry (usually the number of rows).
!arg noc
    Third header entry (usually the number of columns).
!return
    Pointer to open file, or |NULL| on error.
!description
    This function opens a data file for input and writes the the file header.
    If the file does not exist, a new file is created. If the file exists it
    is overwritten.
!seealso FfReadHeader
-------------------------------------------------------------------------- }}}*/

FILE *FfWriteHeader(const char *name, int field, int nor, int noc)

{
    FILE *fd;
    long header[3];

    /* Open the file
       ------------- */
    fd = SysFopen(name,FM_CREATE);
    if (fd == NULL) 
	return NULL;

    /* Write the file header
       --------------------- */
    header[0] = (long) field;
    header[1] = (long) nor;
    header[2]= (long) noc;
    if (SysWriteLong(fd,header,3) != 3)
    {
	MTX_ERROR1("%s: Error writing file header",name);
	fclose(fd);
	return NULL;
    }

    return fd;
}

/*{{{
!definesection kernel.ff.io "File i/o"
    * Binary file format
    Binary data files contain a sequence of objects. Each object consists of a 
    12-byte header, followed by the data part. The data is stored in the
    internal 
    format (see section \ref{GdSection:kernel.ff.ff}) with the following 
    differences:
    - Packed rows (vectors) are truncated to their actual length. For
      example, a row with 17 entries over GF(125) on a machine with
      |sizeof(long)==4| would occupy 20 bytes in memory but only
      17 bytes in a data file.
    - All integers, including the file header, are stored as 32-bit
      numbers with least significant byte first.
    The object header consists of three integers. The first one, $T$, specifies
    which kind of object follows. The two other values, $R$ and $C$, carry
    additional information about the object. Their exact meaning depends on
    the value of $T$.
    \begin{center}\begin{tabular}{|l|l|l|l|}
    \hline
    $T$ & Type of object	    & $R$ 		& $C$ \\
    \hline
    > 1 & Matrix over $\GF{T}$  & no.\ of rows	& no.\ of columns \\
    -1  & Permutation	    & no.\ of points	& no.\ of permutations \\
    -2  & Polynomial	    & field		& degree \\
    -8  & Integer matrix	    & no.\ of rows	& no.\ of columns \\
    \hline
    \end{tabular}
    \end{center}

}}}*/
