/* ============================= C MeatAxe ==================================
   File:        $Id: mat2vec.c,v 1.1.1.1 2000/01/07 16:42:26 mringe Exp $
   Comment:     Convert matrix to vector.
   --------------------------------------------------------------------------
   (C) Copyright 1999 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"


/* --------------------------------------------------------------------------
   Local data
   -------------------------------------------------------------------------- */

MTX_DEFINE_FILE_INFO 


/*{{{ --------------------------------------------------------------------------
!definesection algo.tp
    These function are used for calculations with tensor products.
-------------------------------------------------------------------------- }}}*/


/*{{{ --------------------------------------------------------------------------
!section algo.tp
!function MatrixToVector "Convert matrix to vector"
!arg mat
    Matrix to convert.
!arg vecs
    Destination of the vector.
!arg n
    Row number where the vector is stored.
!return
    $0$ on success, $-1$ on error.
!description
    This function converts a matrix into a row vector by concatenating
    the rows of the matrix. If |mat| is an $r$ by $c$ matrix, the resulting
    vector has $rc$ entries. Instead of allocating a new buffer for the
    result, |MatToVec| expects a pointer to a matrix, |vecs|, and puts 
    the vector into the |n|-th row of this matrix. Of course, |vecs| must
    be over the smae field as |mat|, have at least $n+1$ rows and $rc$
    columns.
!seealso VectorToMatrix
-------------------------------------------------------------------------- }}}*/

int MatrixToVector(const Matrix_t *mat, Matrix_t *vecs, int n)

{
    int i;

    /* Check arguments.
       ---------------- */
    if (!MatIsValid(mat))
    {
	MTX_ERROR1("mat: %E",MTX_ERR_BADARG);
	return -1;
    }
    if (!MatIsValid(vecs))
    {
	MTX_ERROR1("vecs: %E",MTX_ERR_BADARG);
	return -1;
    }
    if (   mat->Nor * mat->Noc != vecs->Noc
	|| mat->Field != vecs->Field)
    {
	MTX_ERROR1("mat and vecs: %E",MTX_ERR_INCOMPAT);
	return -1;
    }
    if (n < 0 || n >= vecs->Nor)
    {
	MTX_ERROR3("n=%d (nor=%d): %E",n,vecs->Nor,MTX_ERR_BADARG);
	return -1;
    }

    /* Convert the matrix.
       ------------------- */
    for (i = 0; i < mat->Nor; ++i)
    {
	if (MatCopyRegion(vecs,n,i*mat->Noc, mat,i,0,1,mat->Noc) != 0)
	{
	    MTX_ERROR("Copying failed");
	    return -1;
	}
    }

    return 0;
}


