/* ============================= C MeatAxe ==================================
   File:        $Id: matcmp.c,v 1.1.1.1 2000/01/07 16:42:26 mringe Exp $
   Comment:     Compare matrices.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <stdlib.h>

MTX_DEFINE_FILE_INFO




/*{{{ --------------------------------------------------------------------------
!section obj.mat
!function MatCompare "Compare two matrices"
!arg a
    Pointer to the first matrix.
!arg b
    Pointer to the second matrix.
!return
    $0$, if the matrices are equal, a nonzero value otherwise, $-1$ on error.
!description
    This function compares two matrices. If the matrices are equal, the return
    value is $0$. Otherwise the return value is positive, if |a| is `greater'
    than  |b|, and negative, if |a| is `less' than |b|. The ordering of 
    matrices is defined as follows:
    \begin{enumerate}
    \item If the matrices are over different fields, the matrix over the
	smaller field is smaller.
    \item Otherwise, if the matrices have different number of columns, the
	matrix with less columns is smaller.
    \item Otherwise, if the matrices have different number of rows, the
	matrix with less rows is smaller.
    \item Otherwise, the relation is determined by the return value of
	|FfCmpRow()| on the first row that is not equal in both matrices.
    \end{enumerate}
    Note that the return value $-1$ does not necessarily mean that an error 
    occured.
!seealso
-------------------------------------------------------------------------- }}}*/

int MatCompare(const Matrix_t *a, const Matrix_t *b)

{
    int i;

    /* Check the arguments
       ------------------ */
    if (!MatIsValid(a) || !MatIsValid(b))
    {
	MTX_ERROR1("%E",MTX_ERR_BADARG);
	return -1;
    }

    /* Compare fields and dimensions
       ----------------------------- */
    if ((i = a->Field - b->Field) != 0)
	return i;
    if ((i = a->Noc - b->Noc) != 0)
	return i;
    if ((i = a->Nor - b->Nor) != 0)
	return i;

    /* Compare the entries row by row. We do not use memcmp on the
       whole matrix because we must ignore padding bytes.
       ----------------------------------------------------------- */
    FfSetField(a->Field);
    FfSetNoc(a->Noc);
    for (i = 0; i < a->Nor; ++i)
    {
	PTR pa = MatGetPtr(a,i);
	PTR pb = MatGetPtr(b,i);
	int diff = FfCmpRows(pa,pb);
	if (diff != 0)
	    return diff;
    }

    /* The matrices are equal!
       ----------------------- */
    return 0;
}

