/* ============================= C MeatAxe ==================================
   File:        $Id: matcopy.c,v 1.1.1.1 2000/01/07 16:42:26 mringe Exp $
   Comment:     Cut a rectangular piece out of a matrix.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"


   
/* --------------------------------------------------------------------------
   Local data
   -------------------------------------------------------------------------- */

MTX_DEFINE_FILE_INFO


/*{{{ --------------------------------------------------------------------------
!section obj.mat
!function MatCopyRegion "Copy a rectanglular region of a matrix"
!arg dest
    Pointer to the destination matrix.
!arg destrow
    Destination row.
!arg destcol
    Destination column.
!arg src
    Pointer to the source matrix.
!arg row1
    First row in region.
!arg col1
    First column in region.
!arg nrows
    Number of rows to cut. $-1$ means as many rows as possible.
!arg ncols
    Number of columns to cut. $-1$ means as many columns as possible.
!return
    Returns $0$ on success, or $-1$ on error.
!description
    This function copies a rectangular region of a matrix into another matrix.
    |row1| and |col1| define the upper left corner of the region. |nrows| and 
    |ncols| are the height (number of 
    rows) and width (number of columns) of the region. Both |nrows| and |ncols| 
    can be $-1$. In this case the region extends up to the last row or last 
    column, respectivly. |destrow| and |destcol| specify where the rectangle
    is placed inside |dest|. As uusual, all row and column indexes start with
    0.

    |src| and |dest| muts be matrices over the same field. Both source and 
    destination rectangle must not exceed the size of the matrix. In particular,
    it is not possible to extend |dest|'s size with this function.
!seealso MatCut
-------------------------------------------------------------------------- }}}*/

int MatCopyRegion(Matrix_t *dest, int destrow, int destcol, 
    const Matrix_t *src, int row1, int col1, int nrows, int ncols)

{
    PTR s, d;
    int i, k;

    /* Check the arguments
       ------------------- */
    if (!MatIsValid(src) || !MatIsValid(dest))
	return -1;
    if (src->Field != dest->Field)
	return MTX_ERROR1("%E",MTX_ERR_INCOMPAT), -1;
    if (nrows == -1)
	nrows = src->Nor - row1;
    if (ncols == -1)
	ncols = src->Noc - col1;
    if (row1 < 0 || nrows < 0 || row1 + nrows > src->Nor)
    {
	MTX_ERROR("Source row index out of range");
	return -1;
    }
    if (col1 < 0 || ncols < 0 || col1 + ncols > src->Noc)
    {
	MTX_ERROR("Source column index out of range");
	return -1;
    }
    if (destrow < 0 || destrow + nrows > dest->Nor)
    {
	MTX_ERROR("Destination row index out of range");
	return -1;
    }
    if (destcol < 0 || destcol + ncols > dest->Noc)
    {
	MTX_ERROR("Destination column index out of range");
	return -1;
    }

    /* Initialize data pointers
       ------------------------ */
    FfSetField(src->Field);
#ifdef PARANOID
    s = row1 < src->Nor ? MatGetPtr(src,row1) : NULL;
	d = destrow < dest->Nor ? MatGetPtr(dest,destrow) : NULL;
#else
    s = MatGetPtr(src,row1);
    d = MatGetPtr(dest,destrow);
#endif

    /* Copy the rectangle
       ------------------ */
    for (i = row1; i < row1 + nrows; ++i)
    {
	for (k = col1; k < col1 + ncols; ++k)
	{
#ifdef PARANOID
	    FEL f;
	    FfSetNoc(src->Noc);
	    f = FfExtract(s,k);
	    FfSetNoc(dest->Noc);
	    FfInsert(d,destcol+k-col1,f);
#else
	    FfInsert(d,destcol+k-col1,FfExtract(s,k));
#endif
	}
	/*s = FfGetPtr(s,1,src->Noc);*/
	s = (PTR)((char *)s + src->RowSize);
	/*d = FfGetPtr(d,1,dest->Noc);*/
	d = (PTR)((char *)d + dest->RowSize);
    }

    Mat_DeletePivotTable(dest);

    return 0;
}




