/* ============================= C MeatAxe ==================================
   File:        $Id: matcore.c,v 1.1.1.1 2000/01/07 16:42:26 mringe Exp $
   Comment:     Basic matrix functions.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <string.h>

   
/* --------------------------------------------------------------------------
   Local data
   -------------------------------------------------------------------------- */

MTX_DEFINE_FILE_INFO

#define MAT_MAGIC 0x6233af91

   
   

/*{{{ --------------------------------------------------------------------------
!definesection obj.mat
\index{Matrix_t@{\tt Matrix\protect\_t} data type}
In the {\MeatAxe}, matrices over finite fields are represented by |Matrix_t|
structures. Besides the marks, each matrix carries the field order, the 
number of rows and the number of columns. There is no global field order
or row length as at the kernel layer.

Matrices can be created in many ways, including (i) calling |MatAlloc()|, 
(ii) making a copy of an existing matrix with |MatDup()|, and (iii) by reading 
a matrix from a data file with |MatRead()| or |MatLoad()|.
The application is responsible for releasing matrices which are no longer
needed. Matrices can consume large amounts of memory, so it always a good
idea to delete a matrix as early as possible. There is only one possibility 
of deleting a matrix: calling |MatFree()|.
-------------------------------------------------------------------------- }}}*/


/*{{{ --------------------------------------------------------------------------
!section obj.mat
!function MatIsValid "Check a matrix"
!synopsis 
    int MatIsValid(const Matrix_t *mat);
!arg mat
    Pointer to the matrix.
!return
    1 if |mat| points to a valid matrix, 0 otherwise.
!description
    This function checks if the argument |mat| is a pointer to a valid
    matrix. If the matrix is o.k., the function returns 1. Otherwise, an error 
    is signaled and, if the error handler does not terminate the program,
    the function returns 0.
!seealso Matrix_t
-------------------------------------------------------------------------- }}}*/

int MatIsValid(const Matrix_t *mat)

{
    if (mat == NULL)
    {
	MTX_ERROR("NULL matrix");
	return 0;
    }
    if (mat->Magic != MAT_MAGIC || mat->Field < 2 || mat->Nor < 0 || 
	mat->Noc < 0)
    {
	MTX_ERROR3("Invalid matrix (field=%d, nor=%d, noc=%d)",mat->Field,
	    mat->Nor,mat->Noc);
	return 0;
    }
    return 1;

}




/*{{{ --------------------------------------------------------------------------
!function MatAlloc "Create a new matrix"
!synopsis 
    Matrix_t *MatAlloc(int field, int nor, int noc);
!arg field
    Field order.
!arg nor
    Number of rows.
!arg noc
    Number of columns.
!return
    Pointer to the new matrix or |NULL| on error.
!description
    This function creates a new matrix with |nor| rows and |noc|
    columns over GF(\verb"field"). The return value is a pointer to the 
    allocated |Matrix_t| structure, or |NULL| if there is not enough memory. 

    To destroy a matrix, use |MatFree()|, {\em not} \verb"SysFree()".
!seealso MatFree Matrix_t
-------------------------------------------------------------------------- }}}*/

Matrix_t *MatAlloc(int field, int nor, int noc)

{
    Matrix_t *m;

    MTX_VERIFY(field >= 2);
    MTX_VERIFY(nor >= 0);
    MTX_VERIFY(noc >= 0);

    /* Allocate a new Matrix_t structure
       --------------------------------- */
    m = ALLOC(Matrix_t);
    if (m == NULL)
    {
	MTX_ERROR("Cannot allocate Matrix_t structure");
	return NULL;
    }

    /* Initialize the data structure
       ----------------------------- */
    if (FfSetField(field) != 0)
    {
	MTX_ERROR1("Cannot select field GF(%d)",field);
	SysFree(m);
	return NULL;
    }
    FfSetNoc(noc);
    m->Magic = MAT_MAGIC;
    m->Field = field;
    m->Nor = nor;
    m->Noc = noc;
    m->PivotTable = NULL;
    m->Data = FfAlloc(nor);
    m->RowSize = FfCurrentRowSize;
    if (m->Data == NULL)
    {
	SysFree(m);
	MTX_ERROR("Cannot allocate matrix data");
	return NULL;
    }
    return m;
}




/*{{{ --------------------------------------------------------------------------
!function MatGetPtr "Pointer to a row of a matrix"
!arg mat
    Pointer to the matrix.
!arg row
    Row index.
!return
    Pointer to the selected row, or |NULL| on error.
!description
    This function returns a pointer to the row number |row| of a matrix.
    Row numbers start from 0. The current row size is not changed.
!seealso
-------------------------------------------------------------------------- }}}*/

PTR MatGetPtr(const Matrix_t *mat, int row)

{
#ifdef DEBUG
    if (!MatIsValid(mat))
	return NULL;
#ifdef PARANOID
    if (row < 0 || row >= mat->Nor)
#else
    if (row < 0 || row > mat->Nor + 5)
#endif
    {
	MTX_ERROR2("row=%d: %E",row,MTX_ERR_BADARG);
	return NULL;
    }
#endif
    return (PTR)((char *) mat->Data + mat->RowSize * row);
}



/*{{{ --------------------------------------------------------------------------
!function Mat_DeletePivotTable "Delete the pivot table a matrix"
!synopsis 
    void Mat_DeletePivotTable(Matrix_t *m);
!arg mat
    Pointer to the matrix.
!description
    This function deletes the pivot table associated with a matrix. It is used
    internally. Applications should never call this function directly.
!seealso
-------------------------------------------------------------------------- }}}*/

void Mat_DeletePivotTable(Matrix_t *mat)

{
    if (mat->PivotTable != NULL)
    {
	SysFree(mat->PivotTable);
	mat->PivotTable = NULL;
    }
}



/*{{{ --------------------------------------------------------------------------
!function MatFree "Delete a matrix"
!synopsis 
    int MatFree(Matrix_t *m);
!arg mat
    Pointer to the matrix.
!return
    $0$ on success, $-1$ on error.
!description
    This function frees a matrix which has beed created by |MatAlloc()|. This
    implies freeing the internal data buffers as well as the |Matrix_t| 
    structure itself.
!seealso MatAlloc Matrix_t
-------------------------------------------------------------------------- }}}*/

int MatFree(Matrix_t *mat)

{
    if (!MatIsValid(mat))
	return -1;
    Mat_DeletePivotTable(mat);
    if (mat->Data != NULL)
	SysFree(mat->Data);
    memset(mat,0,sizeof(Matrix_t));
    SysFree(mat);
    return 0;
}


