/* ============================= C MeatAxe ==================================
   File:        $Id: matech.c,v 1.1.1.1 2000/01/07 16:42:26 mringe Exp $
   Comment:     Reduce a matrix to semi echelon form.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <stdlib.h>


/*{{{ --------------------------------------------------------------------------
!definesection obj.mat
\paragraph{Echelon form and pivot tables}
A matrix $(a_{ij})$ is said to be in {\em echelon form}\index{echelon form} 
if the following conditions are satisfied:
\begin{enumerate}
\item Each row has a first non-zero element (the {\em pivot element}).
\item If $a_{ij}$ is a pivot element, then all elements below are zero,
    i.e., $a_{kj}=0$ for all $k>i$.
\end{enumerate}
If a matrix is in echelon form, the column indexes of its pivot elements are 
called the {\em pivot columns} of the matrix. The list of all pivot columns 
is called the {\em pivot table} of the matrix.

A {\MeatAxe} matrix carries its pivot table in the |PivotTable| member. This
is a pointer to an array of integers containing first the pivot columns
and then the non-pivot columns. This, the size of the |PivotTable| is always
|Noc| integers, the first |Nor| elements containing the pivot columns of the
matrix' rows, and the remaining $|Noc|-|Nor|$ elements containing the non-pivot
columns in arbitrary order. Note that for a matrix in echelon form |Nor| is
always less or equal to |Noc|.
-------------------------------------------------------------------------- }}}*/



MTX_DEFINE_FILE_INFO



static int zmkechelon(PTR matrix, int nor, int noc, int *piv, int *ispiv)

{
    PTR x, newrow;
    int i, j, rank;


    /* Initialize the table
       -------------------- */
    for (i = 0; i < noc; ++i)
    {
	piv[i] = i;
	ispiv[i] = 0;
    }

    /* Echelonize the matrix and build the pivot table in <piv>.
       Keep track of assigned pivot columns in <ispiv>.
       --------------------------------------------------------- */
    rank = 0;
    newrow = matrix;
    for (i = 0, x = matrix; i < nor && rank < noc; ++i, FfStepPtr(&x))
    {
        int newpiv;
        FEL f;

        if (rank < i)
            FfCopyRow(newrow,x);
        FfCleanRow(newrow,matrix,rank,piv);
        newpiv = FfFindPivot(newrow,&f);
	MTX_ASSERT(newpiv < noc);
        if (newpiv >= 0)
        {
	    piv[rank] = newpiv;
	    ispiv[newpiv] = 1;
	    ++rank;
            FfStepPtr(&newrow);
        }
    }

    /* Insert the non-pivot columns
       ---------------------------- */
    j = rank;
    for (i = 0; i < noc; ++i)
    {
	if (!ispiv[i])
	    piv[j++] = i;
    }
    MTX_VERIFY(j == noc);

    return rank;
}




/*{{{ --------------------------------------------------------------------------
!section obj.mat
!function MatEchelonize "Reduce to echelon form"
!synopsis 
    int MatEchelonize(Matrix_t *mat);
!arg mat
    Pointer to the matrix.
!return
    Rank of |mat|, or $-1$ on error.
!description
    This function performs a Gaussian elimination on the matrix |mat|. On
    return, |mat| is in semi echelon form and a pivot table has been 
    attatched to the matrix. If the rank of |mat| was smaller than the number 
    of rows, some rows are removed during the process. This function can also 
    be used to rebuild the pivot table after the matrix has been modified.
!seealso
-------------------------------------------------------------------------- }}}*/

int MatEchelonize(Matrix_t *mat)

{
    int rank;
    int *newtab;
    static int *is_pivot = NULL;
    static int maxnoc = -1;

    /* Check the argument
       ------------------ */
    if (!MatIsValid(mat))
	return -1;

    /* Re-allocate the pivot table. This is not really necessary, since
       |Noc| should never change without releasing the pivot table, but
       this would be a really nasty bug....
       ----------------------------------------------------------------- */
    newtab = NREALLOC(mat->PivotTable,int,mat->Noc);
    if (newtab == NULL)
    {
	MTX_ERROR1("Cannot allocate pivot table (size %d)",mat->Noc);
	return -1;
    }
    mat->PivotTable = newtab;
    if (mat->Noc > maxnoc)
    {
	int *new_is_piv = NREALLOC(is_pivot,int,mat->Noc);
	if (new_is_piv == NULL)
	{
	    MTX_ERROR("Cannot allocate temporary table");
	    return -1;
	}
	is_pivot = new_is_piv;
	maxnoc = mat->Noc;
    }

    /* Build the pivot table
       --------------------- */
    FfSetField(mat->Field);
    FfSetNoc(mat->Noc);
    rank = zmkechelon(mat->Data,mat->Nor,mat->Noc,mat->PivotTable,is_pivot);

    /* If the rank is less than the number of rows, remove null rows
       ------------------------------------------------------------- */
    if (rank != mat->Nor)
    {
	mat->Nor = rank;
	mat->Data = (PTR) SysRealloc(mat->Data,FfCurrentRowSize * rank);
    }

    return rank;
}





/*{{{ --------------------------------------------------------------------------
!function MatNullity "Nullity of a matrix"
!synopsis 
    long MatNullity(const Matrix_t *mat);
!arg mat
    Pointer to the matrix.
!return
    Nullity of |mat|, or $-1$ on error.
!description
    This function calculates the dimension of the null-space of a matrix.
    Unline |MatNullity__()|, this function does not modify the matrix.
!seealso @nullsp
-------------------------------------------------------------------------- }}}*/

long MatNullity(const Matrix_t *mat)

{
    return MatNullity__(MatDup(mat));
}



/*{{{ --------------------------------------------------------------------------
!function MatNullity__ "Nullity of a matrix"
!synopsis 
    long MatNullity__(Matrix_t *mat);
!arg mat
    Pointer to the matrix.
!return
    Nullity of |mat|, or $-1$ on error.
!description
    This function calculates the dimension of the null-space of a matrix
    and deletes the matrix.
!seealso @nullsp
-------------------------------------------------------------------------- }}}*/

long MatNullity__(Matrix_t *mat)

{
    long nul;
    MatEchelonize(mat);
    nul = mat->Noc - mat->Nor;
    MatFree(mat);
    return nul;
}


