/* ============================= C MeatAxe ==================================
   File:        $Id: os.c,v 1.2 2000/09/02 15:32:32 mringe Exp $
   Comment:     OS dependent stuff
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */



/* --------------------------------------------------------------------------
   These sybols can be defined:

   OS_NO_CPU_TIME .......... no CPU times are available
   OS_TIMES_AND_SYSCONF	.... use times() and sysconf() instead of getrusage()
   OS_NO_ITIMER ............ no interval timers
   -------------------------------------------------------------------------- */


/* --------------------------------------------------------------------------
   Include files
   -------------------------------------------------------------------------- */

#include "meataxe.h"

#if defined (_WIN32)

#pragma warning(disable: 4514)
#pragma warning(disable: 4201 4214 4115)
#include <windows.h>
#pragma warning(default: 4201 4214 4115)
#include <direct.h>

#else

#include <sys/stat.h>
#include <sys/time.h>
#include <sys/times.h>
#include <signal.h>
#include <unistd.h>
#include <sys/resource.h>

#endif


#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <stdarg.h>


/* ------------------------------------------------------------------
   Add some missing declarations
   ------------------------------------------------------------------ */


/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

#if defined(OS_NO_CPU_TIME)
time_t zinittime = 0;
#endif


MTX_DEFINE_FILE_INFO

/* ------------------------------------------------------------------
   fopen() modes
   ------------------------------------------------------------------ */

static char *fmodes[7] = { NULL,"rt","wt","at","rb","wb","ab" };



/*{{{ --------------------------------------------------------------------------
!section kernel.os
!function SysInit "Initialize OS dependent stuff"
!synopsis 
    void SysInit();
!description
    This function is called during library initialization. It performs any
    OS-dependent actions. Applications never call this function directly.
-------------------------------------------------------------------------- }}}*/

void SysInit()

{
#if defined(OS_NO_CPU_TIME)
    zinittime = time(NULL);
#endif
}




/*{{{ --------------------------------------------------------------------------
!section kernel.os.time
!function SysTimeUsed "CPU time"
!synopsis 
    long SysTimeUsed();
!return
    CPU time.
!description
    This function returns the CPU time used by the calling process in 
    units of $1/10$ seconds.
!seealso SysSetTimeLimit
-------------------------------------------------------------------------- }}}*/

long SysTimeUsed(void)

{

#if defined(_WIN32)

    FILETIME cr, ex, krnl, usr;
    LARGE_INTEGER k, u;
    GetProcessTimes(GetCurrentProcess(),&cr,&ex,&krnl,&usr);
    memcpy(&k,&krnl,sizeof(k));
    memcpy(&u,&usr,sizeof(u));
    return (long)((k.QuadPart + u.QuadPart) / 1000000);

#elif defined(OS_TIMES_AND_SYSCONF)

    struct tms t;
    static long clk_tck = 0;
    if (clk_tck == 0) 
	clk_tck = sysconf(_SC_CLK_TCK);
    times(&t);
    return ((long)((t.tms_utime + t.tms_stime) * 10 / clk_tck ));

#elif defined(OS_NO_CPU_TIME)

    return (time(NULL) - zinittime) * 10;

#else

    static struct rusage ru;
    getrusage(RUSAGE_SELF,&ru);
    return (ru.ru_utime.tv_sec * 10 + ru.ru_utime.tv_usec / 100000);

#endif


}




/*{{{ --------------------------------------------------------------------------
!section kernel.os.time
!function SysSetTimeLimit "Set CPU time limit"
!synopsis 
    void SysSetTimeLimit(long nsecs);
!arg nsecs
    CPU time limit in seconds.
!description
    This function sets a CPU time limit for the calling process. If the
    process does not exit within the specified time, it it killed.

    {\em Note:} This function is not available on all platforms.
!seealso SysTimeUsed
-------------------------------------------------------------------------- }}}*/

#ifdef _WIN32

DWORD CALLBACK Killer(void *x)
{
    DWORD ms = (long) x;
    Sleep(ms);
    MTX_ERROR1("%E",MTX_ERR_GAME_OVER);
    return 0;
}

void SysSetTimeLimit(long nsecs)
{
    DWORD id;
    CreateThread(NULL,0,Killer,(void *)(nsecs * 1000),0,&id);
}

#elif !defined(OS_NO_ITIMER)

void vtalarm(int i)

{
    MTX_ERROR1("%E",MTX_ERR_GAME_OVER);
}

void SysSetTimeLimit(long nsecs)

{
    struct itimerval tv;

    tv.it_interval.tv_sec = 0;
    tv.it_interval.tv_usec = 0;
    tv.it_value.tv_sec = nsecs;
    tv.it_value.tv_usec = 0;
    setitimer(ITIMER_VIRTUAL,&tv,NULL);
    signal(SIGVTALRM,vtalarm);
}

#else /* No interval timer, sorry... */

void SysSetTimeLimit(long nsecs)

{
}

#endif




/*{{{ --------------------------------------------------------------------------
!section kernel.os.io
!function SysFopen "Open file"
!synopsis 
    FILE *SysFopen(const char *name, int mode);
!arg name
    File name.
!arg mode
    Open mode and additional flags (see below).
!return
    Pointer to an open file or |NULL| on error.
!description
    This function opens a file. |mode| must be any of |FM_READ| (open 
    for reading), |FM_CREAT| (create a new file) or |FM_APPEND| (append 
    to existing file or create a new file).  Additional flags may be or'ed
    to the mode:
    \begin{center}
    \begin{tabular}{|l|p{0.7\linewidth}|}
    \hline
    Mode & Meaning \\
    \hline
    |FM_LIB| & If the file does not exist in the current directory,
	    look in the library directory. The library directory is
	    defined either by the environment variable \verb"MTXLIB" or
	    at compile-time by the symbol \verb"MTXLIB".\\
    \hline
    |FM_TEXT| & Open in text mode. This flag must be used on some
	    systems (e.g., MS-DOS) to open text files. By default, files
	    are assumed to contain binary data.\\
    \hline
    |FM_NOERROR| & Do not generate an error if the file does not exist.\\
    \hline
    \end{tabular}
    \end{center}
    The return value is a pointer to the open file or NULL on error.
!seealso FfReadHeader FfWriteHeader
-------------------------------------------------------------------------- }}}*/

FILE *SysFopen(const char *name, int mode)

{
    char buf[300];
    int m;
    FILE *f;

    m = mode & 0x0F;			/* Append, read or create? */
    if ((mode & FM_TEXT) == 0) m += 3;	/* Binary mode */
    if (m < 1 || m > 6 || (mode & 0x0F) == 0)
    {
	MTX_ERROR1("Invalid file mode %d",mode);
	return NULL;
    }
    f = fopen(name,fmodes[m]);
    if (f != NULL) 
	return f;

    /* Search library directory
       ------------------------ */
    if ((mode & FM_LIB) != 0) 
    {
	strcpy(buf,MtxLibDir);
	strcat(buf,"/");
	strcat(buf,name);
	f = fopen(buf,fmodes[m]);
    }

    /* Error handling
       -------------- */
    if (f == NULL && (mode & FM_NOERROR) == 0)
	MTX_ERROR1("%s: %S",name);

    return f;
}



/*{{{ --------------------------------------------------------------------------
!section kernel.os.io
!function SysFseek "Set file pointer"
!synopsis
    int SysFseek(FILE *file, long pos);
!arg file
    Pointer to |FILE| structure.
!arg pos
    New position of file pointer.
!return
    The function returns 0 on success. Any other value indicates an error.
!description
    This function sets the file pointer to a given position. if |pos| is greater
    or equal to zero, it is interpreted as an absolute position (relative to
    start of file). If |pos| is negative, the file pointer is moved to the end 
    of file.
!seealso SysFseekRelative FfSeekRow 
-------------------------------------------------------------------------- }}}*/

int SysFseek(FILE *file, long pos)

{
    if (pos < 0)
	return fseek(file,(long) 0,SEEK_END);
    else
	return fseek(file,pos,SEEK_SET);
}




/*{{{ --------------------------------------------------------------------------
!section kernel.os.io
!function SysRemoveFile "Remove a file"
!arg name
    File name.
!return
    $0$ on success, $-1$ on error.
!description
    This function deletes a file. On a UNIX system, |SysRemoveFile()|
    is equivalent to |remove()|. If the file to be deleted does not exist,
    or cannot be removed for some other reason, a run-time error error is 
    generated.
!seealso 
-------------------------------------------------------------------------- }}}*/

int SysRemoveFile(const char *name)

{
    if (remove(name) != 0)
    {
	MTX_ERROR1("Cannot remove file '%s'",name);
	return -1;
    }
    return 0;
}




/*{{{ --------------------------------------------------------------------------
!section kernel.os.io
!function SysRemoveDirectory "Remove a directory"
!arg name
    Directory name.
!return
    $0$ on success, $-1$ on error.
!description
    This function removes a directory. If the directory is not empty, or
    cannot be removed for some other reason, a run-time error is
    generated and the fuction returns $-1$.
!seealso  SysCreateDirectory
-------------------------------------------------------------------------- }}}*/

int SysRemoveDirectory(const char *name)

{
    if (rmdir(name) != 0)
    {
	MTX_ERROR1("Cannot remove directory '%s'",name);
	return -1;
    }
    return 0;
}





/*{{{ --------------------------------------------------------------------------
!section kernel.os.io
!function SysCreateDirectory "Create a directory"
!arg name
    Directory name.
!return
    $0$ on success, $-1$ on error.
!description
    This function creates a new directory. If the directory cannot be 
    created for some reason, a run-time error is generated and the fuction 
    returns $-1$.
!seealso  SysRemoveDirectory
-------------------------------------------------------------------------- }}}*/

int SysCreateDirectory(const char *name)

{
#ifdef _WIN32
    if (mkdir(name) != 0)
#else
    if (mkdir(name,0755) != 0)
#endif
    {
	MTX_ERROR1("Cannot create directory '%s'",name);
	return -1;
    }
    return 0;
}


/*{{{ --------------------------------------------------------------------------
!section kernel.os.io
!function SysFseekRelative "Set file pointer"
!arg file
    Pointer to |FILE| structure.
!arg distance
    Number of bytes to move.
!return
    The function returns 0 on success. Any other value indicates an error.
!description
    This function moves the file pointer by a given number of bytes.
    |distance can be positive or negative.
!seealso SysFseek FfSeekRow
-------------------------------------------------------------------------- }}}*/

int SysFseekRelative(FILE *file, int distance)

{
    return fseek(file,distance,SEEK_CUR);
}



/*{{{ --------------------------------------------------------------------------
!section kernel.os.mem
!function SysMalloc "Allocate memory"
!synopsis 
    void *SysMalloc(size_t nbytes);
!arg nbytes
    Size of memory block to allocate.
!return
    Pointer to memory block or |NULL| on error.
!description
   This function works like |malloc()|. Unlike some implementations of 
   |malloc()|, it returns a valid pointer if |nbytes| is zero.
!seealso @OsMem
-------------------------------------------------------------------------- }}}*/

void *SysMalloc(size_t nbytes)

{
    void *x;

    if (nbytes == 0) 
	nbytes = 1;
    x = malloc(nbytes);
    if (x == NULL)
	MTX_ERROR1("Cannot allocate %d bytes: %S",(int) nbytes);
    /*
    memset(x,0,nbytes);
    */
    return x;
}



/*{{{ --------------------------------------------------------------------------
!section kernel.os.mem
!function SysRealloc "Resize memory block"
!synopsis 
    void *SysRealloc(size_t nbytes);
!arg buf
    Pointer to memory block.
!arg nbytes
    Desired size.
!return
    Pointer to resized memory block or NULL on error.
!description
   This function works like |relloc()| but handles zero-length blocks 
   correctly.
!seealso @OsMem
-------------------------------------------------------------------------- }}}*/

void *SysRealloc(void *buf, size_t nbytes)

{
    void *x;

    if (nbytes == 0) 
	nbytes = 1;
    x = realloc(buf,nbytes);
    if (x == NULL)
	MTX_ERROR1("Cannot reallocate %d bytes: %S",(int) nbytes);
    return x;
}



/*{{{ --------------------------------------------------------------------------
!section kernel.os.mem
!function SysFree "Free memory block"
!synopsis 
    void *SysFree(void *x);
!arg x
    Pointer to the memory block.
!description
   This function works like |free()| but checks if the argument is not
   NULL. Otherwise, an appropriate error message is generated.
!seealso @OsMem
-------------------------------------------------------------------------- }}}*/

void SysFree(void *x)

{
    if (x == NULL)
	MTX_ERROR("Attempt to free() NULL pointer");
    else
	free(x);
}



/*{{{ --------------------------------------------------------------------------
!section kernel.os
!function SysGetPid "Get process id"
!return
    Process id.
!description
    This function returns a number which uniquely identifies the calling
    process on the local system. The exact meaning of this number depends
    on the operating system. In an UNIX environment, it is the process id
    (PID).
!seealso 
-------------------------------------------------------------------------- }}}*/

int SysGetPid()

{
    int pid;
#ifdef _WIN32
    pid = (int) GetCurrentProcessId();
#else
    pid = (int) getpid();
#endif
    return pid;
}




/*{{{ --------------------------------------------------------------------------
!definesection kernel.os
The {\MeatAxe} is written for a UNIX-like operating environment and uses
many functions of the standard C library. To make the {\MeatAxe} more 
portable between different operating systems, some C library and system 
calls are accessed through wrapper functions. These wrapper functions have
names that begin with 'Sys'. For example |SysFree()| is the wrapper
function for |free()|.
-------------------------------------------------------------------------- }}}*/
