/* ============================= C MeatAxe ==================================
   File:        $Id: permcore.c,v 1.2 2000/09/02 15:32:32 mringe Exp $
   Comment:     Basic permutation functions.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <string.h>

   
/* --------------------------------------------------------------------------
   Local data
   -------------------------------------------------------------------------- */

MTX_DEFINE_FILE_INFO

#define PERM_MAGIC 0x30f8326b

   

/*{{{ --------------------------------------------------------------------------
!definesection obj.perm
\index{permutations}
Permutations operate on the set $\{0,1,\ldots,n-1\}$, where $n$ is the
degree. A permutation is represented by the |Perm_t| data structure.
This structure also carries the degree of the permutation. This is 
important because you cannot, for example, multiply permutations of 
different degrees.
-------------------------------------------------------------------------- }}}*/






/*{{{ --------------------------------------------------------------------------
!section obj.perm
!function PermIsValid "Check a permutation"
!arg p
    Pointer to the permutation.
!return
    1 if |p| points to a valid permutation, 0 otherwise.
!description
    This function checks if the argument, |p|, is a pointer to a valid
    permutation. If the permutation is o.k., the function return 1. 
    Otherwise, an error is signaled and, if the error handler does not 
    terminate the program, the function returns 0.
!seealso Perm_t
-------------------------------------------------------------------------- }}}*/

int PermIsValid(const Perm_t *p)

{
    int i, deg;
    long *x;

    if (p == NULL)
    {
	MTX_ERROR("NULL permutation");
	return 0;
    }
    deg = p->Degree;
    if (p->Magic != PERM_MAGIC || deg < 0 || p->Data == NULL)
    {
	MTX_ERROR2("Invalid permutation (magic=%d, deg=%d)",
	    p->Magic,deg);
	return 0;
    }
    for (i = p->Degree, x = p->Data; i > 0; --i, ++x)
    {
	if (*x < 0 || *x >= deg)
	{
	    MTX_ERROR2("Invalid value %d in permutation (deg = %d)",
		(int) *x,deg);
	    return 0;
	}
    }

    return 1;

}


/*{{{ --------------------------------------------------------------------------
!function PermAlloc "Allocate a permutation"
!synopsis 
    Perm_t *PermAlloc(int deg);
!arg deg
    Degree.
!return
    Pointer to the new permutation, or |NULL| on error.
!description
    This function creates a permutation of the degree |deg|. It returns 
    a pointer to the new permutation |NULL| on error. The new permutation
    is initialized to the identity.
!seealso PermFree
-------------------------------------------------------------------------- }}}*/

Perm_t *PermAlloc(int deg)

{
    Perm_t *p;
    int i;

    if (deg < 0)
    {
	MTX_ERROR2("deg=%d: %E",deg,MTX_ERR_BADARG);
	return NULL;
    }

    p = ALLOC(Perm_t);
    if (p == NULL)
    {
	MTX_ERROR("Cannot allocate Perm_t structure");
	return NULL;
    }
    p->Magic = PERM_MAGIC;
    p->Degree = deg;
    p->Data = NALLOC(long,deg);
    if (p->Data == NULL)
    {
	SysFree(p);
	MTX_ERROR("Cannot allocate permutation data");
	return NULL;
    }
    for (i = 0; i < deg; ++i)
	p->Data[i] = i;
    return p;
}



/*{{{ --------------------------------------------------------------------------
!function PermFree "Free a permutation"
!section obj.perm
!synopsis 
    int PermFree(Perm_t *p);
!arg p
    Pointer to the permutation.
!return
    The function returns $0$ on success and $-1$ on error.
!description
    This function deletes a permutation and returns the memory to the 
    system. Do not use |SysFree()| on permutations because this would only 
    free the |Perm_t| structure but not the data buffer.
!seealso PermAlloc
-------------------------------------------------------------------------- }}}*/

int PermFree(Perm_t *p)

{
    /* Check arguments
       --------------- */
    if (!PermIsValid(p))
	return -1;
  
    SysFree(p->Data);
    memset(p,0,sizeof(Perm_t));
    SysFree(p);
    return 0;
}
