/* ============================= C MeatAxe ==================================
   File:        $Id: permpwr.c,v 1.1.1.1 2000/01/07 16:42:27 mringe Exp $
   Comment:     Power of a permutation.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
   


/* --------------------------------------------------------------------------
   Local data
   -------------------------------------------------------------------------- */

MTX_DEFINE_FILE_INFO


/*{{{ --------------------------------------------------------------------------
!section obj.perm
!function PermPower "Power of a permutation"
!arg p
    Pointer to the permutation.
!arg n
    Exponent.
!return
    |n|-th power of |p| or |NULL| on error.
!description
    This function calculates the \verb"n"-th power of a permutation.
    $n$ must be positive or 0.
    It allocates a new permutation on each call, leaving the original
    permutation intact. The caller is responsible for deleting the 
    result when it is no longer needed.
!seealso
-------------------------------------------------------------------------- }}}*/

Perm_t *PermPower(const Perm_t *p, int n)

{
    Perm_t *q;
    long *xp;
    long *xq;
    long i, k, l;

    /* Check arguments 
       --------------- */
    if (!PermIsValid(p))
	return NULL;
    if (n < 0)
    {
	MTX_ERROR1("Invalid exponent %ld < 0",n);
	return NULL;
    }

    /* Allocate a new permutation for the result
       ----------------------------------------- */
    q = PermAlloc(p->Degree);
    if (q == NULL)
	return NULL;
    xp = p->Data;
    xq = q->Data;
    
    /* Calculate the n-th power
       ------------------------ */
    for (i = 0; i < p->Degree; ++i)
    {	
	for (k = i, l = n; l > 0; --l)
	    k = xp[k];
	xq[i] = k;
    }
    return q;
}




