/* ============================= C MeatAxe ==================================
   File:        $Id: rdcfgen.c,v 1.1.1.1 2000/01/07 16:42:27 mringe Exp $
   Comment:     Read generators for constituents.
   --------------------------------------------------------------------------
   (C) Copyright 1997  Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"

MTX_DEFINE_FILE_INFO


/*{{{ --------------------------------------------------------------------------
!section lattice
!function Lat_ReadCfGens "Load a constituent"
!synopsis 
    MatRep_t *Lat_ReadCfGens(Lat_Info *info, int cf, int flags);
!arg info
    Pointer to the lattice information structure.
!arg cf
    Number of constituent to load.
!arg flags
    Optional flags (see below).
!return
    Pointer to a matrix representation of the specified constituent,
    or |NULL| on error.
!description
     This function reads the generators of one constituent and optionally
     performs some basic operations on the genrators (invert, transpose).
     |flags| may be any combination of the followong values:
     \begin{center}
     \begin{tabular}{|l|p{0.7\linewidth}|}
     \hline
	 |LAT_RG_STD| & Read generators in standard basis (file XY.std.N).
			These files are created by pwkond. Default is to
			read the generators in 'random' basis as they are
			produced by chop.\\
	 \hline
	 |LAT_RG_INVERT| & Invert the generators\\
	 \hline
	 |LAT_RG_TRANSPOSE|  & Transpose the generators\\
	 \hline
     \end{tabular}
     \end{center}

!seealso 
-------------------------------------------------------------------------- }}}*/

MatRep_t *Lat_ReadCfGens(Lat_Info *info, int cf, int flags)

{
    MatRep_t *rep;
    char fn[LAT_MAXBASENAME + 20];
    int i;

    /* Check the arguments
       ------------------- */
    if (info == NULL)
    {
	MTX_ERROR1("info: %E",MTX_ERR_BADARG);
	return NULL;
    }
    if (cf < 0 || cf >= info->NCf)
    {
	MTX_ERROR1("cf: %E",MTX_ERR_BADARG);
	return NULL;
    }

    /* Make file name
       -------------- */
    if (flags & LAT_RG_STD)
	sprintf(fn,"%s%s.std.%%d",info->BaseName,Lat_CfName(info,cf));
    else
	sprintf(fn,"%s%s.%%d",info->BaseName,Lat_CfName(info,cf));

    /* Load the representation
       ----------------------- */
    rep = MrLoad(fn,info->NGen);
    if (rep == NULL)
	return NULL;

    /* Apply modifications
       ------------------- */
    for (i = 0; i < rep->NGen; ++i)
    {
	if (flags & LAT_RG_INVERT)
	{
	    Matrix_t *mat2 = MatInverse(rep->Gen[i]);
	    if (mat2 == NULL)
	    {
		MTX_ERROR1("Cannot transpose generator %d",i);
		MrFree(rep);
		return NULL;
	    }
	    MatFree(rep->Gen[i]);
	    rep->Gen[i] = mat2;
	}
	if (flags & LAT_RG_TRANSPOSE)
	{
	    Matrix_t *mat2 = MatTransposed(rep->Gen[i]);
	    if (mat2 == NULL)
	    {
		MTX_ERROR1("Cannot invert generator %d",i);
		MrFree(rep);
		return NULL;
	    }
	    MatFree(rep->Gen[i]);
	    rep->Gen[i] = mat2;
	}
    }
    return rep;
}

