/* ============================= C MeatAxe ==================================
   File:        $Id: saction.c,v 1.1.1.1 2000/01/07 16:42:27 mringe Exp $
   Comment:     Action on a subspace.
   --------------------------------------------------------------------------
   (C) Copyright 1999 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <stdlib.h>



MTX_DEFINE_FILE_INFO 



/*{{{ --------------------------------------------------------------------------
!section algo.split
!function SAction "Action on a subspace"
!arg subspace
    Pointer to an invariant subspace.
!arg gen
    Matrix operating on the subspace.
!return
    Action of the generator on the subspace, or |NULL| on error.
!description
    Given a subspace $U\leq F^n$ and a matrix $a\in F^{n\times n}$ that maps 
    $U$ into $U$, this function calculates the action of the matrix on the 
    subspace.
    As input, the function expects a basis of the subspace in |subspace|,
    which must be in chelon form, and the matrix operating on the
    subspace in |gen|. The result is a square matrix with $\dim(V)$ rows 
    containing the image of the basis vectors under $a$, expressed in the 
    given basis.
    
    Before calculating the action, |SAction()| checks if the arguments
    are valid matrices, and if they are compatible. Both |subspace|
    and |gen| must be over the same field, have the same number of
    columns, and |gen| must be square.

    Note that |SAction()| is intended to be used with single matrices.
    If you want to compute the action of a whole set of matrices on the 
    same subspace, |Split()| may be more suitable.
!seealso QAction Split
-------------------------------------------------------------------------- }}}*/

Matrix_t *SAction(const Matrix_t *subspace, const Matrix_t *gen)

{
    int dim, sdim, i;
    PTR tmp;
    Matrix_t *action;

   /* Check arguments.
      ---------------- */
    if (!MatIsValid(subspace) || !MatIsValid(gen))
	return NULL;
    if (subspace->Noc != gen->Nor)
    {
	MTX_ERROR1("subspace and gen: %E",MTX_ERR_INCOMPAT);
	return NULL;
    }
    if (gen->Nor != gen->Noc)
    {
	MTX_ERROR1("gen: %E",MTX_ERR_NOTSQUARE);
	return NULL;
    }

   /* Set up internal variables.
      -------------------------- */
    dim = subspace->Noc;
    sdim = subspace->Nor;
    FfSetField(subspace->Field);
    action = MatAlloc(FfOrder,sdim,sdim);
    FfSetNoc(dim);
    tmp = FfAlloc(1);

    /* Calaculate the action.
       ---------------------- */
    for (i = 0; i < subspace->Nor; ++i)
    {
	PTR xi = MatGetPtr(subspace,i);
	PTR yi = MatGetPtr(action,i);
	FEL f;

	/* Calculate the image of the <i>-th row of <subspace>.
	   ---------------------------------------------------- */
	FfMapRow(xi,gen->Data,dim,tmp);

	/* Clean the image with the subspace and store coefficients.
	   --------------------------------------------------------- */
	FfCleanRow2(tmp,subspace->Data,sdim,subspace->PivotTable,yi);
	if (FfFindPivot(tmp,&f) >= 0)
	    MTX_ERROR("Split(): Subspace not invariant");
    }

    /* Clean up and return the result.
       ------------------------------- */
    SysFree(tmp);
    return action;
}


