/* ============================= C MeatAxe ==================================
   File:        $Id: setcore.c,v 1.1.1.1 2000/01/07 16:42:27 mringe Exp $
   Comment:     Basic set functions.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <string.h>

   
/* --------------------------------------------------------------------------
   Local data
   -------------------------------------------------------------------------- */

MTX_DEFINE_FILE_INFO

static const int InitialSize = 10;
static const unsigned long SetMagic = 0xEF452338;




/*{{{ --------------------------------------------------------------------------
!definesection obj.set
    The {\MeatAxe} has two data types that represent subsets of the natural
    numbers, |Set_t| and |BitString_t|.
-------------------------------------------------------------------------- }}}*/



/*{{{ --------------------------------------------------------------------------
!section obj.set
!function SetIsValid "Check a set"
!synopsis 
    int SetIsValid(const Set_t *s);
!arg s
    Pointer to the set.
!return
    1 if |s| points to a valid set, 0 otherwise.
!description
    This function checks if the argument, |s|, is a pointer to a valid
    set. If the set is o.k., the function returns 1. 
    Otherwise, an error is signaled and, if the error handler does not 
    terminate the program, the function returns 0.
!seealso Perm_t
-------------------------------------------------------------------------- }}}*/

int SetIsValid(const Set_t *s)

{
    if (s == NULL)
    {
	MTX_ERROR("NULL set");
	return 0;
    }
    if (s->Magic != SetMagic || s->Size < 0 || s->BufSize < s->Size)
    {
	MTX_ERROR3("Invalid set (Magic=%d, Size=%d, BufSize=%d)",
	    (int)s->Magic,s->Size,s->BufSize);
	return 0;
    }
    if (s->Data == NULL)
    {
	MTX_ERROR("Data==NULL in set");
	return 0;
    }
    return 1;
}




/*{{{ --------------------------------------------------------------------------
!function SetAlloc "Create a new set"
!synopsis 
    Set_t *SetAlloc();
!return
    Pointer to the new set or |NULL| on error.
!description
    This function creates a new, empty set. To destroy a set, 
    use \verb"SetFree()", {\em not} \verb"SysFree()".
!seealso SetFree Set_t
-------------------------------------------------------------------------- }}}*/

Set_t *SetAlloc()

{
    Set_t *x;

    x = ALLOC(Set_t);
    if (x == NULL)
    {
 	MTX_ERROR("Cannot allocate set");
	return NULL;
    }
    x->Size = 0;
    x->BufSize = InitialSize;
    x->Data = NALLOC(long,InitialSize);
    if (x->Data == NULL)
    {
	SysFree(x);
 	MTX_ERROR("Cannot allocate set data");
	return NULL;
    }
    x->Magic = SetMagic;
    return x;
}




/*{{{ --------------------------------------------------------------------------
!function SetFree "Destroy a set"
!synopsis 
    int SetFree(Set_t *x);
!arg x
    Pointer to the set.
!return
    $0$ on success, $-1$ on error.
!description
    This function frees an integer set. The argument must be a \verb"Set_t" 
    structure which has previously been allocated with \verb"SetAlloc()". 
!seealso SetAlloc Set_t
-------------------------------------------------------------------------- }}}*/

int SetFree(Set_t *x)

{
    if (!SetIsValid(x))
	return -1;
    SysFree(x->Data);
    memset(x,0,sizeof(Set_t));
    SysFree(x);
    return 0;
}




/*{{{ --------------------------------------------------------------------------
!function SetDup "Duplicate a set"
!synopsis 
    Set_t *SetDup(const Set_t *s);
!arg s
    Pointer to the set.
!return
    Pointer to a copy of the set, or |NULL| on error.
!description
    This function creates a copy of an existing set.
!seealso SetAlloc
-------------------------------------------------------------------------- }}}*/

Set_t *SetDup(const Set_t *s)

{
    Set_t *x;

    if (!SetIsValid(s))
	return NULL;
    x = ALLOC(Set_t);
    if (x == NULL)
    {
 	MTX_ERROR("Cannot allocate set");
	return NULL;
    }

    x->Size = s->Size;
    x->BufSize = s->Size;
    x->Data = NALLOC(long,x->BufSize);
    if (x->Data == NULL)
    {
	SysFree(x);
 	MTX_ERROR("Cannot allocate set data");
	return NULL;
    }
    memcpy(x->Data,s->Data,sizeof(x->Data[0]) * s->Size);
    x->Magic = SetMagic;
    return x;
}

