/* ============================= C MeatAxe ==================================
   File:        $Id: sumint.c,v 1.2 2000/03/18 13:58:43 mringe Exp $
   Comment:     Sum and intersection of vector spaces (Zassenhaus algorithm)
   --------------------------------------------------------------------------
   (C) Copyright 1999 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <string.h>

MTX_DEFINE_FILE_INFO


/*{{{ --------------------------------------------------------------------------
!section kernel.ff.row
!function FfSumAndIntersection "Sum and intersection of vector spaces"
!arg wrk1
    Work space 1 (see below).
!arg nor1
    Input: Dimension $V$. Output: Dimension of $V+W$.
!arg nor2
    Input: Dimension $W$. Output: Dimension of $V\cap W$.
!arg wrk2
    Work space 2 (see below).
!arg piv
    Pivot table.
!return
    $0$ on success, $-1$ on error.
!index "Zassenhaus algorithm"
!index "Intersection of vector spaces"
!index "Sum of vector spaces"
!description
    Given two vector spaces $V$ and $W$, this function calculates the sum
    and the intersection of the spaces, using the Zassenhaus algorithm.
    It is assumed that the two spaces are given by a generating set of
    vectors. These sets need not be linearly independent.
    Before calling this |SunAndIntersection|, the caller must allocate and
    initialize two workspaces and a pivot table:
    - Both workspaces must have $n_1+n_2$ rows, where $n_i$ is the number
      of generating vectors for $V$ and $W$, respectively. 
    - Work space 1, |wrk1|, must contain the concatenation of the
      generating sets for $V$ and $W$. Work space 2 need not be initialized.
    - The pivot table, |piv|, must be large enough for at least 
      $n_1+n_2$ entries. It need not be initialized.
    - The variables |*nor1| and |*nor2| must contain the numbers $n_1$
      and $n_2$, respectively.
    On return, |*nor1| contains the dimension of $V+W$, and |*nor2| contains
    the dimension of $V\cap W$. The first |*nor1| rows of |wrk1| contain a
    basis of $V+W$, and a basis of $V\cap W$ can be found in |wrk2| starting
    at position |*nor1|. Both bases are in echelon form, and the |piv| contains
    the pivot table for the bases.
!seealso 
-------------------------------------------------------------------------- }}}*/

int FfSumAndIntersection(PTR wrk1, int *nor1, int *nor2, PTR wrk2, int *piv)

{
    int dim1 = *nor1, dim2 = *nor2;
    int i, k, sumdim;
    PTR x1, x2, y1, y2, sec;


    /* Check the arguments.
       -------------------- */
    if (wrk1 == NULL || nor1 == NULL || nor2== NULL
	|| wrk2 == NULL || piv == NULL)
    {
	MTX_ERROR1("%E",MTX_ERR_BADARG);
	return -1;
    }

    /* Set up the workspace 2. Initially, it contains a copy of V1.
       ------------------------------------------------------------ */
    for (x2 = wrk2, i = 0; i < dim1 + dim2; ++i, FfStepPtr(&x2))
	FfMulRow(x2,FF_ZERO);
    memcpy(wrk2,wrk1,dim1 * FfCurrentRowSize);

    /* Step 1: Echelonize workspace 1, repeating 
       all operations on workspace 2.
       ----------------------------------------- */
    x1 = y1 = wrk1;
    x2 = y2 = wrk2;
    k = 0;
    for (i = 0; i < dim1 + dim2; ++i, FfStepPtr(&x1), FfStepPtr(&x2))
    {
	FEL f;
	int p;
	FfCleanRowAndRepeat(x1,wrk1,k,piv,x2,wrk2);
	if ((p = FfFindPivot(x1,&f)) < 0)
	    continue;	/* Null row - ignore */
	if (k < i)
	{
	    FfSwapRows(y1,x1);
	    FfSwapRows(y2,x2);
	}
	piv[k++] = p;
	FfStepPtr(&y1);
	FfStepPtr(&y2);
    }
    sumdim = k;	/* Dimension of V + W */

    /* Step 2: Echelonize the basis of the intersection.
       ------------------------------------------------- */
    sec = x2 = y2;
    for (i = sumdim; i < dim1 + dim2; ++i, FfStepPtr(&x2))
    {
	FEL f;
	int p;
	FfCleanRow(x2,sec,k - sumdim,piv + sumdim);
	if ((p = FfFindPivot(x2,&f)) < 0)
	    continue;
	if (i > k)
	    FfCopyRow(y2,x2);
	piv[k++] = p;
	FfStepPtr(&y2);
    }

    *nor1 = sumdim;	    /* Dimension of U + W */
    *nor2 = k - sumdim;	    /* Dimension of the intersection */

    return 0;
}


