/* ============================= C MeatAxe ==================================
   File:        $Id: wgen.c,v 1.1.1.1 2000/01/07 16:42:27 mringe Exp $
   Comment:     Word generator.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <string.h>

/*{{{ --------------------------------------------------------------------------
!definesection algo.wgen
Given a finitely generated matrix algebra $A$, the word generator produces 
`random' elements of $A$, i.e., words in the generators.
-------------------------------------------------------------------------- }}}*/



/* ------------------------------------------------------------------
   Local data
   ------------------------------------------------------------------ */

MTX_DEFINE_FILE_INFO

#define MINLEN	5
#define MAXLEN	8

static int B0Tab[8][MAXLEN+1] = {
    {0,-1,-1,-1,-1}, {1,-1,-1,-1,-1}, {2,3,-1,-1,-1}, {5,4,-1,-1,-1},
    {7,9,-1,-1,-1}, {6,11,13,-1,-1}, {17,8,1,-1,-1}, {19,10,21,23,-1}
};

static int B1Tab[8][MAXLEN+1] = {
    {0,1,2,-1,-1}, {4,3,5,6,-1},{8,10,7,-1,-1},{12,9,14,11,-1},
    {13,16,15,18,-1}, {17,20,22,19,-1},{24,21,23,25,26,-1},{27,29,28,31,33,-1}
};

static int BitTab[238] = 
{
    0x07,0x27,0xD9,0xDB,0xDF,0xF9,0xE0,0x03,0x05,0x06,0x09,0x0A,0x0B,0x0C,
    0x0D,0x0E,0x0F,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x1A,0x1B,
    0x1C,0x1D,0x1E,0x1F,0x21,0x22,0x23,0x24,0x25,0x26,0x28,0x29,0x2A,0x2B,
    0x2C,0x2D,0x2E,0x2F,0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,
    0x3A,0x3B,0x3C,0x3D,0x3E,0x3F,0x41,0x42,0x43,0x44,0x45,0x46,0x47,0x48,
    0x49,0x4A,0x4B,0x4C,0x4D,0x4E,0x4F,0x50,0x51,0x52,0x53,0x54,0x55,0x56,
    0x57,0x58,0x59,0x5A,0x5B,0x5C,0x5D,0x5E,0x5F,0x60,0x61,0x62,0x63,0x64,
    0x65,0x66,0x67,0x68,0x69,0x6A,0x6B,0x6C,0x6D,0x6E,0x6F,0x70,0x71,0x72,
    0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7A,0x7B,0x7C,0x7D,0x7E,0x81,0x82,
    0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,0x8F,0x90,
    0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9A,0x9B,0x9C,0x9D,0x9E,
    0x9F,0xA0,0xA1,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAA,0xAB,0xAC,
    0xAD,0xAE,0xAF,0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBA,
    0xBB,0xBC,0xBD,0xBE,0xC0,0xC1,0xC2,0xC3,0xC4,0xC5,0xC6,0xC7,0xC8,0xC9,
    0xCA,0xCB,0xCC,0xCD,0xCE,0xCF,0xD0,0xD1,0xD2,0xD3,0xD4,0xD5,0xD6,0xD7,
    0xD8,0xDA,0xDC,0xDD,0xDE,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,
    0xEA,0xEB,0xEC,0xED,0xEE,0xF0,0xF1,0xF2,0xF3,0xF4,0xF6,0xF8,0xFA,0xFC 
};


#define RND(x) (((x) * 214013L + 2531011L) & 0xFFFFFFFF)
#define Mod ((1 << (MAXLEN+1)) - 2)


static int CalcLen(int n2)
{
    if (n2 < 200)
	return 5;
    if (n2 < 2000)
	return 6;
    if (n2 < 20000)
	return 7;
    return MAXLEN;
}


static void MakeBufX2(int n2, int buf[8][MAXLEN+1])
{
    unsigned long r = n2;
    int i;
    int m[8];
    int len, mod;

    len = CalcLen(n2);
    mod = (1 << (len+1)) - 2;

    for (i = 0; i < 8; )
    {
	int mm, k;
	r = RND(r);
	mm = (r >> 16) % mod + 2;
	for (k = 0; k < i && m[k] != mm; ++k);
	if (k >= i)
	    m[i++] = mm;
    }
    for (i = 0; i < 8; ++i)
    {
	unsigned long x = m[i];
	int k = len;
	int mask = 1 << len;
	while ((x & mask) == 0)
	{
	    buf[i][k] = -1;
	    mask >>= 1;
	    --k;
	};
        buf[i][k] = -1;
	mask >>= 1;
	--k;
	while (k >= 0)
	{
	    int gg = (int)((r >> 16) * 2);
	    r = RND(r);
	    buf[i][k] = ((x & mask) == 0) ? gg : gg + 1;
	    mask >>= 1;
	    --k;
	}
    }
}


static void MakeBuf(int n2, int ngen, int buf[8][MAXLEN+1])

{
    int pos;

    if (n2 == 0)
	memcpy(buf,B0Tab,sizeof(B0Tab));
    else if (n2 == 1)
	memcpy(buf,B1Tab,sizeof(B0Tab));
    else
	MakeBufX2(n2,buf);
    for (pos = 0; pos < 8; ++pos)
    {
	int *x;
	for (x = buf[pos]; *x >= 0; ++x)
	    *x %= ngen;
    }
}


static const int *B(int n2, int pos, int ngen)

{
    static int buf[8][MAXLEN+1];
    static int lastn2 = -1;
 
    if (n2 != lastn2)
    {
	MakeBuf(n2,ngen,buf);
	lastn2 = n2;
    }
    return buf[pos];
}


static char *BasisName(int n2, int pos, int ngen)
{
    static char buf[MAXLEN+1];
    char *c = buf;
    const int *x = B(n2,pos,ngen);
    while (*x >= 0)
    {
	int i;
	*c++ = (char)(*x++ + 'a');
	for (i = 1; i < 9 && *x == x[-1]; ++i, ++x);
	if (i > 1)
	    *c++ = (char)(i + '0');
    }
    *c = 0;
    return buf;
}

static void GenBasis(WgData_t *b, int n2, int pos)
{
    const int *x;
    Matrix_t *buf = NULL;

    if (b->Basis[pos] != NULL)
	MatFree(b->Basis[pos]);
    for (x = B(n2,pos,b->Rep->NGen); *x >= 0; ++x)
    {
	MTX_ASSERT(*x >= 0 && *x < b->Rep->NGen);
	if (buf == NULL)
	    buf = MatDup(b->Rep->Gen[*x]);
	else
	    MatMul(buf,b->Rep->Gen[*x]);
    }
    MTX_VERIFY(buf != NULL);
    b->Basis[pos] = buf;
    b->N2[pos] = n2;
}




/*{{{ --------------------------------------------------------------------------
!section algo.wgen
!function WgSymbolicName "Symbolic name of a word"
!synopsis 
    const char *WgSymbolicName(const WgData_t *b, long n);
!arg b
    Pointer to word generator data.
!arg n
    Word number.
!return
    Symbolic name of the word |n|.
!description
    This function returns a symbolic representation of the word number |n| as
    a polynomial in the generators. The generators are named a, b, c\ldots
    The return value is a pointer to a static buffer which is overwritten on 
    each call.
!seealso @wgen
-------------------------------------------------------------------------- }}}*/

const char *WgSymbolicName(const WgData_t *b, long n)

{
    int i;
    static char name[8 * (MAXLEN+1)+1];
    char *c = name;
    long n1, n2;

    MTX_VERIFY(n > 0);
    --n;
    n1 = BitTab[(int)(n % 238)];
    n2 = (int)(n / 238);

    for (i = 0; i < 8 && n1 != 0; ++i, n1 >>= 1)
    {
	const char *bn;
	if ((n1 % 2) == 0)
	    continue;
	bn = BasisName(n2,i,b->Rep->NGen);
	if (c != name)
	    *c++ = '+';
	strcpy(c,bn);
	while (*c != 0) ++c;
    }
    return name;
}



/*{{{ --------------------------------------------------------------------------
!section algo.wgen
!function WgMakeWord "Make a word"
!synopsis 
    Matrix_t *WgMakeWord(WgData_t *b, long n);
!arg b
    Pointer to word generator data.
!arg n
    Word number.
!return
    Matrix representation of word |n|, or |NULL| on error.
!description
    This function calculates a `random' element in a matrix algebra.
    The algebra, i.e., a set of generating matrices, must be specified with
    |WgAlloc()|, and |b| is the pointer returned by |WgAlloc()|.
    |n| is the number of the element to calculate. Refer to section
    \ref{GdSection:algo.wgen} for more details.

    Here is an example demonstrating the usage of the word generator:
    \begin{verbatim}
	MatRep_t *rep;
	wgdata_t *wg;
	Matrix_t *word;
	long nul;
	...
	wg = WgAlloc(rep);
	word = WgMakeWord(wg,1833);
	nul = MatNullity__(word);
	printf("Word 1833 has nullity %ld\n",nul);
	WgFree(wg);
    \end{verbatim}
    This example initializes the word generator with three matrices,
    calculates the word number 1833 and prints out its nullity.
!seealso @wgen
-------------------------------------------------------------------------- }}}*/

Matrix_t *WgMakeWord(WgData_t *b, long n)

{
    Matrix_t *w = NULL;
    int n1, n2;
    int i;

    MTX_VERIFY(n > 0);
    --n;
    n1 = BitTab[(int)(n % 238)];
    n2 = (int)(n / 238);
    for (i = 0; i < 8 && n1 != 0; ++i, n1 >>= 1)
    {
	if ((n1 % 2) == 0)
	    continue;
	if (b->N2[i] != n2)
	    GenBasis(b,n2,i);
	if (w == NULL)
	    w = MatDup(b->Basis[i]);
	else
	    MatAdd(w,b->Basis[i]);
    }
    return w;
}




static int CheckArgs(const MatRep_t *rep)

{
    if (!MrIsValid(rep))
    {
	MTX_ERROR1("rep: %E",MTX_ERR_BADARG);
	return -1;
    }
    if (rep->NGen < 1) 
    {
	MTX_ERROR1("Invalid number of generators (%d)",rep->NGen);
	return -1;
    }
    return 0;
}



/*{{{ --------------------------------------------------------------------------
!section algo.wgen
!function WgAlloc "Initialize the word generator"
!synopsis 
    WgData_t *WgAlloc(const MatRep_t *rep);
!arg rep
    Pointer to the matrix representation.
!return
    Pointer to a new word generator data structure, or |NULL| on error.
!description
    This function initializes the word generator for a given matrix
    representation |rep|. There must be at least one generator.
    On success, |WgInit()| returns a pointer to an internal data
    structure which can be used in subsequent calls to |WgMakeWord()|
    and |WgFree()|. If an error occurs, the return value is NULL.

    Note that the word generator does not make internal copies of the
    generators. The caller must assure that the generators, i.e., the
    |MatRep_t| structure, are not deleted or modified in any way as long 
    as the word generator is used.
!seealso WgFree
-------------------------------------------------------------------------- }}}*/

WgData_t *WgAlloc(const MatRep_t *rep)

{
    int k, fl, dim;
    WgData_t *b;

    /* Check the Arguments
       ------------------- */
    if (CheckArgs(rep) != 0)
	return NULL;
    fl = rep->Gen[0]->Field;
    dim = rep->Gen[0]->Nor;

    /* Create a new data structure
       ---------------------------  */
    if ((b = ALLOC(WgData_t)) == NULL)
    {
	MTX_ERROR("Cannot allocate word generator data: %S");
	return NULL;
    }

    /* Initialize the members
       ---------------------- */
    b->Rep = rep;
    for (k = 0; k < 8; ++k)
    {
	b->Basis[k] = NULL;
	b->N2[k] = -1;
    }

    return b;
}



/*{{{ --------------------------------------------------------------------------
!section algo.wgen
!function WgFree "Terminate the word generator"
!synopsis 
    int WgFree(WgData_t *b);
!arg b
    Pointer to word generator data.
!return
    $0$ on success, $-1$ on error.
!description
    This function terminates the word generator specified by |b| and cleans up 
    internal data structures. Note that the generators that were passed to 
    |WgInit()| are {\em not} freed.
!seealso WgAlloc
-------------------------------------------------------------------------- }}}*/

int WgFree(WgData_t *b)

{
    int k;

    /* Check the handle
       ---------------- */
    if (b == NULL)
    {
	MTX_ERROR1("%E",MTX_ERR_BADARG);
	return -1;
    }

    /* Free all basis matrices 
       ----------------------- */
    for (k = 0; k < 8; ++k)
    {
	if (b->Basis[k] != NULL)
    	    MatFree(b->Basis[k]);
    }
    memset(b,0,sizeof(WgData_t));
    SysFree(b);
    return 0;
}





/*{{{ --------------------------------------------------------------------------
!section algo.wgen
!function WgMakeFingerPrint "Calculate finger print"
!synopsis 
    void WgMakeFingerPrint(wgdata_t *b, int fp[6])
!arg b
    Pointer to word generator data.
!arg fp
    Buffer for the finger print (6 numbers).
!description
    This function calculates the `finger print' of a module, i.e. the nullities
    of the first 6 words. |fp| must be a pointer to an array of at least 6
    long integers.
!seealso @wgen
-------------------------------------------------------------------------- }}}*/

void WgMakeFingerPrint(WgData_t *b, int fp[6])

{	
    int i;
    for (i = 1; i <= 6; ++i)
    	fp[i-1] = MatNullity__(WgMakeWord(b,i));
}


