#include "meataxe.h"


MTX_DEFINE_FILE_INFO

static MtxApplicationInfo_t AppInfo = { 
"zcl", "Clean Matrix", 
"SYNTAX\n"
"    zcl <Subsp> <Mat> <Cleaned mat> <Ops>\n"
"\n"
"OPTIONS\n"
MTX_COMMON_OPTIONS_DESCRIPTION
"\n"
"FILES\n"
"    <Subsp> ................. I The subspace to clean with\n"
"    <Mat> ................... I The matrix to be cleaned\n"
"    <Cleaned> ............... O The cleaned matrix\n"
"    <Ops> ................... O Row operations that were performed\n"
};
static MtxApplication_t *App = NULL;


const char *SpcName = NULL, *MatName = NULL, *OpName = NULL, *ClName = NULL;
static Matrix_t *Space;
static MtxFile_t *Matrix, *Cleaned, *Op;


static int Clean1()

{
    PTR op, row;
    int j;
    int rc = 0;

    /* Check compatibility.
       -------------------- */
    if (Matrix->Noc != Space->Noc || Matrix->Field != Space->Field)
    {
	MTX_ERROR3("%s and %s: %E",SpcName,MatName,MTX_ERR_INCOMPAT);
	return 1;
    }

    /* Build the pivot table.
       ---------------------- */
    if (MatPivotize(Space) != 0)
    {
	MTX_ERROR2("%s: %E",SpcName,MTX_ERR_NOTECH);
	return -1;
    }

    /* Allocate work space.
       -------------------- */
    FfSetNoc(Space->Nor);
    if ((op = FfAlloc(1)) == NULL)
	rc = -1;
    FfSetNoc(Space->Noc);
    if ((row = FfAlloc(1)) == NULL)
	rc = -1;

    /* Process the matrix row by row.
       ------------------------------ */
    for (j = 0; rc == 0 && j < Matrix->Nor; ++j)
    {
	if (MfReadRows(Matrix,row,1) != 1)
	    rc = -1;
	FfCleanRow2(row,Space->Data,Space->Nor,Space->PivotTable,op);
	if (MfWriteRows(Cleaned,row,1) != 1)
	    rc = -1;
	if (MfWriteRows(Op,op,1) != 1)
	    rc = -1;
	FfMulRow(op,FF_ZERO);
    }

    /* Clean up.
       --------- */
    if (op != NULL)
	SysFree(op);
    if (row != NULL)
	SysFree(row);

    return rc;
}



int main(int argc, const char **argv)

{
    int rc;


    if ((App = AppAlloc(&AppInfo,argc,argv)) == NULL)
	return -1;
    if (AppGetArguments(App,4,4) < 0)
	return -1;
    SpcName = App->ArgV[0];
    MatName = App->ArgV[1];
    ClName = App->ArgV[2];
    OpName = App->ArgV[3];

    if ((Space = MatLoad(SpcName)) == NULL)
	rc = -1;
    if ((Matrix = MfOpen(MatName)) == NULL)
	rc = -1;
    if ((Cleaned = MfCreate(ClName,FfOrder,Matrix->Nor,Matrix->Noc)) == NULL)
	rc = -1;
    if ((Op = MfCreate(OpName,FfOrder,Matrix->Nor,Space->Nor)) == NULL)
	rc = -1;

    /* Clean.
       ------ */
    rc = Clean1();

    /* Clean up.
       --------- */
    if (Space != NULL) MatFree(Space);
    if (Matrix != NULL) MfClose(Matrix);
    if (Cleaned != NULL) MfClose(Cleaned);
    if (Op != NULL) MfClose(Op);
    if (App != NULL) AppFree(App);

    return rc;
}


/*{{{ --------------------------------------------------------------------------
!section apps.general
!program zcl "Clean a matrix by a space"
!synopsis 
    zcl <Subsp> <Mat> <CleanedMat> <RowOps>
!arg Subsp
    The subspace (matrix to clean with).
!arg Mat
    The matrix to be cleaned.
!arg CleanedMat
    The cleaned matrix.
!arg RowOps
    Operations perfoned during cleaning.
!description
This program reads in two matrices from <Space> and <Matrix>,
and `cleans' <Matrix> with <Space>. It puts out two matrices
to <CleanedMat> and <RowOps> such that
$<Matrix> = <RowOps>\cdot<Space>+<CleanedMat>$
and <CleanedMat> has as many zero columns as possible. The two input 
matrices must be over the same field and have the same number of columns. 
Also, <Space> must be in echelon form. If it is not, it will be 
reduced internally before cleaning the matrix, i.e., the equation above
still holds if <Space> is replaced by the echelon form of <Space>.

One use of this program is to calculate the action of a generator
on an invariant subspace: Take the subspace in echelon form (as it
is on output from ZSP), multiply it by a generator. Feed the original
echelon form into <Space> and the result of the multiplication into
<Matrix>. <CleanedMat> should be zero, and <RowOps> is the
action of the generator on the invariant subspace.
The action on the quotient of a given subspace can be calculated with
the ZQT program (see section \ref{sec:zqd}).

!implementation
The first matrix, <Space> is hold completely in memory. The second 
matrix, <Matrix> is then processed a row at a time. Row operations 
are performed to clear out the pivot points of the input row using the 
rows of the first input matrix. A row describing what was done is written 
out to <RowOps>, and the remnant (clean) row is output to 
<CleanedMat>.

!seealso 
-------------------------------------------------------------------------- }}}*/

