/* ============================= C MeatAxe ==================================
   File:        $Id: znu.c,v 1.3 2002/02/16 13:02:08 mringe Exp $
   Comment:     This program calculates the null space of a matrix.
   --------------------------------------------------------------------------
   (C) Copyright 1999 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */



#include "meataxe.h"
#include <stdlib.h>



/* ------------------------------------------------------------------
   Variables
   ------------------------------------------------------------------ */

MTX_DEFINE_FILE_INFO

static MtxApplicationInfo_t AppInfo = { 
"znu", "Matrix Null-Space", 
"SYNTAX\n"
"    znu [-GQVn] <Matrix> [<NullSpace>]\n"
"\n"
"ARGUMENTS\n"
"    <Matrix> ................ Input file name\n"
"    <Nullspace> ............. Output file name\n"
"OPTIONS\n"
MTX_COMMON_OPTIONS_DESCRIPTION
"    -G, --gap ............... GAP output\n"
"    -n, --no-echelon ........ Do no convert the null-space to echelong form\n"
"\n"
};

static MtxApplication_t *App = NULL;
static int opt_G = 0;
static int opt_n = 0;				/* -n: no echelon form */
static const char *matname = NULL;
static const char *nspname = NULL;
static Matrix_t *Matrix = NULL;



static int Init(int argc, const char **argv)
{
    /* Process command line options.
       ----------------------------- */
    App = AppAlloc(&AppInfo,argc,argv);
    if (App == NULL)
	return -1;
    opt_G = AppGetOption(App,"-G --gap");
    opt_n = AppGetOption(App,"-n --no-echelon");
    if (opt_G)
	MtxMessageLevel = -100;

    /* Process arguments.
       ------------------ */
    if (AppGetArguments(App,1,2) < 0)
	return -1;
    matname = App->ArgV[0];
    if (App->ArgC > 1)
	nspname = App->ArgV[1];

    /* Read the matrix.
       ---------------- */
    MESSAGE(1,("Reading %s\n",matname));
    Matrix = MatLoad(matname);
    if (Matrix == NULL)
	return 1;

    return 0;
}



/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, const char **argv)

{   
    int nspdim;

    if (Init(argc,argv) != 0)
    {
	MTX_ERROR("Initialization failed");
	return 1;
    }

    if (nspname != NULL)
    {
	Matrix_t *null_space = MatNullSpace_(Matrix,opt_n ? 1 : 0);
	if (null_space == NULL)
	    return 1;
        MESSAGE(1,("Writing null-space to %s\n",nspname));
	MatSave(null_space,nspname);
	nspdim = null_space->Nor;
    }
    else
    {
	MatEchelonize(Matrix);
	nspdim = Matrix->Noc - Matrix->Nor;
    }


    if (opt_G)
    {
	printf("MeatAxe.Nullity := %d;\n",nspdim);
	fflush(stdout);
    }
    else
        MESSAGE(0,("NULLITY %d\n",nspdim));


    AppFree(App);
    return 0;
}


/*{{{ --------------------------------------------------------------------------
!section apps.general
!program znu "Matrix null-space"
!index null-space
!index kernel
!synopsis 
    znu [-G] [-Q] [-V] <Matrix> [<NullSpace>]
!arg <Matrix>
    Input file: The matrix.
!arg <NullSpace>
    Output file: The null-space of {\it Matrix}.
!option -Q
    Quiet, no messages.
!option -V
    Verbose, more messages.
!option "-G, --gap"
    GAP output format.
!description
This program reads in a matrix from <Matrix> and outputs a basis
for its null-space in semi-echelon form to <Nullspace>. If the
<Nullspace> argument is omitted the null-space is not written
out, but its dimension is still displayed.

Notice that the input matrix does not need to be square. The
matrix is read in, and the program generates the n times n
identity matrix in memory where n is the number of rows. It then
proceeds to do row operations to the input matrix, and the same
operations to the other one. The input matrix is reduced to
semi-echelon form, and whenever a zero vector is made, the
corresponding row of the other matrix is marked for output.
The null-space is always reduced to semi-echelon form.
!seealso 
-------------------------------------------------------------------------- }}}*/
