/* ============================= C MeatAxe ==================================
   File:        $Id: zpc.c,v 1.1.1.1 2000/01/07 16:42:27 mringe Exp $
   Comment:     Permutation chop.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include <string.h>
#include <stdlib.h>
#include "meataxe.h"


/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

MTX_DEFINE_FILE_INFO

static MtxApplicationInfo_t AppInfo = {
"zpc", "Permutation Chop",
"SYNTAX\n"
"    zpc [-b] <Perm1> <Perm2> <Seed> <S1> <S2> <Q1> <Q2>\n"
"    zpc [-b] -g <#Perm>[.<#Gen>] <Perm> <Seed> <Sub> <Quot>\n"
"\n"
"OPTIONS\n"
"    -b ...................... Find action on block system\n"
"    -g ...................... Multiple permutation mode\n"
"\n"
"FILES\n"
"    <Perm1,2> ............... I Generators: permutations of equal degree\n"
"    <Seed> .................. I Seed: One point or (with -b) one block\n"
"    <S1,2> .................. O Action on the orbit or block system\n"
"    <Q1,2> .................. O Action on the rermaining points\n"
"\n"
};



static MtxApplication_t *App = NULL;
long *seed;			/* Read from G3 */
int npoints;			/* Degree of input permutations */
int blksize;			/* Block size */
int nblocks;			/* Number of blocks */
long *perm[MAXGEN+1];		/* Generators */
int ngen = 2;			/* Number of generators */
int nperm = 2;			/* Number of permutations to chop */
long *orb;
long *num;
long start;
long orblen;

const char *genname[MAXGEN+1];
const char *subname[MAXGEN+1];
const char *quotname[MAXGEN+1];
const char *seedname;

int opt_b = 0;			/* Option -b (permute blocks) */
int opt_g = 0;			/* Option -g (additional generators) */


static char *helptext[] = {
NULL};





/* ------------------------------------------------------------------
   mkname()
   ------------------------------------------------------------------ */

static char *mkname(const char *basename,int i)

{   
    char *c;

    c = NALLOC(char,strlen(basename)+10);
    sprintf(c,"%s%d",basename,i);
    return c;
}


/* ------------------------------------------------------------------
   setnperm()
   ------------------------------------------------------------------ */

static void setnperm(const char *c)

{
    nperm = -1;
    opt_g = 1;
    if (sscanf(c,"%d.%d",&nperm,&ngen) == 1)
	ngen = nperm;
    if (nperm > MAXGEN || ngen > nperm || ngen < 2)
	MTX_ERROR1("-g: %E",MTX_ERR_OPTION);
    if (ngen != nperm && MSG1)
        MTX_ERROR2("%d generators, %d permutations",ngen,nperm);
}


/* ------------------------------------------------------------------
   parseargs() - Process command line options and arguments
   ------------------------------------------------------------------ */

static int parseargs()

{
    int i;
    const char *c;
    int args_needed;

    opt_b = AppGetOption(App,"-b --block-system");
    c = AppGetTextOption(App,"-g",NULL);

    opt_g = c != NULL;
    if (opt_g)
	setnperm(c);

    /* Get file names
       -------------- */
    args_needed = opt_g ? 4 : 7;
    if (AppGetArguments(App,args_needed,args_needed) < 0)
	return -1;
    if (opt_g)
    {
	seedname = App->ArgV[1];
	for (i = 1; i <= nperm; ++i)
	{	
	    genname[i] = mkname(App->ArgV[0],i);
	    subname[i] = mkname(App->ArgV[2],i);
	    quotname[i] = mkname(App->ArgV[3],i);
	}
    }
    else
    {	
	quotname[2] = App->ArgV[6];
	quotname[1] = App->ArgV[5];
	subname[2] = App->ArgV[4];
	subname[1] = App->ArgV[3];
	seedname = App->ArgV[2];
	genname[2] = App->ArgV[1];
	genname[1] = App->ArgV[0];
    }
    return 0;
}



/* ------------------------------------------------------------------
   init() - Initialize everything, read input files
   ------------------------------------------------------------------ */

static int init(int argc, char **argv)

{
    long *m;
    int i;
    MtxFile_t *f;


    App = AppAlloc(&AppInfo,argc,argv);
    if (App == NULL)
	return -1;
    parseargs(argc, argv);

    /* Read the Permutation
       -------------------- */
    for (i = 1; i <= nperm; ++i)
    {
	if ((f = MfOpen(genname[i])) == NULL)
	    return -1;
	if (f->Field != -1 || f->Noc != 1)
	{
	    MTX_ERROR2("%s: %E",genname[i],MTX_ERR_NOTPERM);
	    return -1;
	}
	if (i == 1)
	    npoints = f->Nor;
	else if (f->Nor != npoints)
	{
	    MTX_ERROR3("%s and %s: %E",genname[1],genname[i],MTX_ERR_INCOMPAT);
	    return -1;
	}
    	m = NALLOC(long,npoints);
	if (m == NULL)
	    return -1;
	if (MfReadLong(f,m,npoints) != npoints)
	    return -1;
	Perm_ConvertOld(m,npoints);
        perm[i] = m;
	MfClose(f);
    }

    /* Seed
       ---- */
    if ((f = MfOpen(seedname)) == NULL)
	return -1;
    if (f->Field != -1) 
    {
	MTX_ERROR3("%s: %E (found type %d, expected -1)",seedname,
	    MTX_ERR_FILEFMT,(int)f->Field);
    }
    if (!opt_b)
    {	
	blksize = 1;
	nblocks = npoints;
    }
    else
    {	
	blksize = f->Nor;
	nblocks = npoints / blksize;
	if (npoints % blksize != 0)
	{
	    MTX_ERROR("BLOCK SIZE DOES NOT DIVIDE DEGREE");
	    return -1;
	}
    }

    seed = NALLOC(long,blksize);
    if (seed == NULL) 
	return -1;
    if (MfReadLong(f,seed,blksize) != blksize)
	return -1;
    MfClose(f);

    /* Allocate tables
       --------------- */
    if ((orb = NALLOC(long,npoints+1)) == NULL 
	|| (num = NALLOC(long,npoints+1)) == NULL
       )
       return -1;
    for (i = 0; i < npoints; ++i)
    {
	orb[i] = -1;
	num[i] = -1;
    }

    return 0;
}


/* ------------------------------------------------------------------
   writeresult() - Calculate action on orbits and write output files.
   ------------------------------------------------------------------ */

static int writeresult()

{   
    int cosize;
    long *s, *p;
    int i;
    long k, im;
    MtxFile_t *f;

    cosize = npoints - orblen;
    if (cosize == 0)
	printf("Transitive on %ld %s\n",orblen/blksize,
		opt_b ? "blocks" : "points");
    else
	printf("Intransitive - 'sub' %ld  'quot' %ld\n",
		orblen/blksize,cosize/blksize);

    /* Calculate action on first orbit
       ------------------------------- */
    if (orblen % blksize != 0) 
    {
	MTX_ERROR2("Invalid block system: orblen=%d, blksize=%d",
	    orblen,blksize);
    }
    s = NALLOC(long,orblen/blksize+1);
    for (i = 1; i <= nperm; ++i)
    {	
	p = perm[i];
	for (k = 0; k < orblen/blksize; ++k)
	    s[k] = -1;
	for (k = 0; k < npoints; ++k)
	{
	    long x, y;
	    if (num[k] >= orblen) continue;
	    x = num[k] / blksize;
	    y = num[p[k]] / blksize;
	    if (s[x] == -1)
		s[x] = y;
	    else
	    {
		if (s[x] != y) 
		    MTX_ERROR("Invalid block system");
	    }
	}
	if ((f = MfCreate(subname[i],-1,orblen/blksize,1)) == NULL)
	    return -1;
	if (MfWriteLong(f,s,orblen/blksize) != (orblen/blksize))
	    return -1;
	MfClose(f);
    }
    free(s);
    if (opt_b || cosize <= 0) 
	return -1;

    /* Calculate action on other orbits (not with -b option)
       ----------------------------------------------------- */
    if ((s = NALLOC(long,cosize+1)) == NULL)
	return -1;
    for (i = 1; i <= nperm; ++i)
    {	
	p = perm[i];
	for (k = 1; k < npoints; ++k)
	{
	    if (num[k] < orblen) 
		continue;
	    im = p[k];
	    s[num[k]-orblen] = num[im]-orblen;
	}
	if ((f = MfCreate(quotname[i],(long)-1,cosize,(long)1)) == NULL)
	    return -1;
	if (MfWriteLong(f,s,cosize) != cosize)
	    return -1;
	MfClose(f);
    }
    free(s);
    return 0;
}


/* ------------------------------------------------------------------
   chop() - Make orbit
   ------------------------------------------------------------------ */

static void chop()

{	
    long level, i, k, newpt;
    int gen;

    /* Set up the orbit to contain only the seed point/block
       ----------------------------------------------------- */
    for (i = 0; i < blksize; ++i)
    {	
	orb[i] = seed[i];
	num[orb[i]] = i;
    }
    orblen = blksize;
    level = 0;

    /* Make orbit
       ---------- */
    while (level < orblen)
    {	/* Apply all generators
           -------------------- */
	for (gen = 1; gen <= ngen; ++gen)
	{	
	    if (num[perm[gen][orb[level]]] == -1)
			/* New point ? */
	    {	
		for (i = 0; i < blksize; ++i)
		{	
		    newpt=perm[gen][orb[level+i]];
		    num[newpt] = orblen+i;
		    orb[orblen+i] = newpt;
		}
		orblen += blksize;
	    }
	}
	level += blksize;
    }

    if (opt_b)	/* No `quotient' when permuting blocks */
	return;

    /* Renumber the remaining points
       ----------------------------- */
    k = orblen;
    for (i = 0; i < npoints; ++i)
    {	
	if (num[i] == -1)
	    num[i] = ++k;
    }
}


/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, char *argv[])

{
    if (init(argc,argv) != 0)
    {
	MTX_ERROR("Initialization failed");
	return 1;
    }

    chop();
    writeresult();
    AppFree(App);
    return EXIT_OK;
}
