/* ============================= C MeatAxe ==================================
   File:        $Id: zsc.c,v 1.3 2001/04/04 17:59:38 mringe Exp $
   Comment:     Spinup with script
   --------------------------------------------------------------------------
   (C) Copyright 1999 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"

#include <string.h>
#include <stdlib.h>



/* --------------------------------------------------------------------------
   Global data
   -------------------------------------------------------------------------- */

MTX_DEFINE_FILE_INFO

static MtxApplicationInfo_t AppInfo = { 
"zsc", "Spin up with script",
"\n"
"SYNTAX\n"
"    zsc [<Options>] <Gen> <Seed> <Op> [<Out>]\n"
"\n"
"ARGUMENTS\n"
"    <Gen> ................... Generator name\n"
"    <Seed> .................. Seed vector(s)\n"
"    <Op> .................... Spin-up script\n"
"    <Out> ................... Result\n"        
"\n"
"OPTIONS\n"
MTX_COMMON_OPTIONS_DESCRIPTION
"    -g <#Gen> ............... Set number of generators.\n"
"\n"
"FILES\n"
"    <Gen>.{1,2...} .......... I  Generators\n"
"    <Seed> .................. I  Seed vectors\n"
"    <Op> .................... I  Spin-up script\n"
"    <Out> ................... I  Output file\n"
};

static MtxApplication_t *App = NULL;

static const char *GenName = NULL;	/* File name of generators */
static const char *SeedName = NULL;	/* Seed vector file name */
static const char *OutName = NULL;	/* Output file name (base name) */
static const char *OpName = NULL;	/* Name of <op> file */

static int fl, nor;		/* Field and size of the generators */
static IntMatrix_t *OpTable;	/* The spin-up script */
static int ngen = 2;		/* Number of generators */
static MatRep_t *Rep;		/* The generators */
static Matrix_t *Seed;		/* Seed vectors */

#define OPVEC(i) OpTable->Data[i * 2]
#define OPGEN(i) OpTable->Data[i * 2 + 1]

/* --------------------------------------------------------------------------
   ReadFiles() - Read input files

   Description:
     This function reads the generators and the script file. It also
     allocates a workspace for spin-up.
   -------------------------------------------------------------------------- */
	     
static int ReadFiles()

{
    int i;

    /* Read the generators.
      --------------------- */
    Rep = MrLoad(GenName,ngen);
    if (Rep == NULL)
	return -1;
    fl = Rep->Gen[0]->Field;
    nor = Rep->Gen[0]->Nor;

    /* Read the script.
       ---------------- */
    MESSAGE(1,("Reading %s\n",OpName));
    OpTable = ImatLoad(OpName);
    if (OpTable == NULL)
	return -1;
    ConvertSpinUpScript(OpTable);   /* 2.3 compatibility */
    if (OpTable->Noc != 2)
    {
	MTX_ERROR1("%s: bad number of columns",OpName);
	return -1;
    }
   /* Das ist doch unsinnig! J.M. */
   /* if (OpTable->Nor != nor)
    {
	MTX_ERROR1("%s: bad number of rows",OpName);
	return -1;
    } */

    /* Check the script for errors
       --------------------------- */
    /* if (OPVEC(0) != 0 || OPGEN(0) != -1) nanana, J.M. */
    if (OPVEC(0) != 1 || OPGEN(0) != -1)
	MTX_ERROR("Illegal skript (does not start with first vector)");
    for (i = 1; i < OpTable->Nor; ++i)
    {
	if (OPGEN(i) == -1)
	    MTX_ERROR("Illegal skript (more than 1 seed vector)");
	if (OPGEN(i) < 0 || OPGEN(i) >= ngen)
	    MTX_ERROR1("Illegal skript (pos %d: generator out of range)",i);
	/* if (OPVEC(i) <= 0 || OPVEC(i) >= i) nanana, J.M. */
	if (OPVEC(i) < 0 || OPVEC(i) >= i)
	    MTX_ERROR1("Illegal skript (pos %d: vector out of range)",i);
    }

    /* Read the seed vectors.
       ---------------------- */
    Seed = MatLoad(SeedName);
    if (Seed == NULL)
	return -1;
    if (Seed->Noc != nor || Seed->Field != fl)
    {
	MTX_ERROR3("%s.1 and %s: %E",GenName,SeedName,MTX_ERR_INCOMPAT);
	return 1;
    }
    MESSAGE(1,("%s: %d seed vectors\n",SeedName,Seed->Nor));
    return 0;
}



/* ------------------------------------------------------------------
   Init() - Program initialization
   ------------------------------------------------------------------ */

static int Init(int argc, const char **argv)

{
    if ((App = AppAlloc(&AppInfo,argc,argv)) == NULL)
	return -1;

    /* Options.
       -------- */
    ngen = AppGetIntOption(App,"-g",2,1,1000);

    /* Arguments.
       ---------- */
    if (AppGetArguments(App,3,4) < 0)
	return -1;
    OpName = App->ArgV[2];
    SeedName = App->ArgV[1];
    GenName = App->ArgV[0];
    OutName = App->ArgC > 3 ? App->ArgV[3] : SeedName;

    /* Read input files.
       ----------------- */
    if (ReadFiles() != 0)
	return -1;

    return 0;
}






/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, const char *argv[])

{
    int seedno;

    if (Init(argc,argv) != 0)
    {
	MTX_ERROR("Initialization failed");
	return 1;
    }

    /* Spinup each seed vector.
       ------------------------ */
    for (seedno = 0; seedno < Seed->Nor; ++seedno)
    {
	char fn[200];
	Matrix_t *seed_vec = MatCutRows(Seed,seedno,1);
	Matrix_t *result = SpinUpWithScript(seed_vec,Rep,OpTable);
        if (result == NULL)
	    return 1;
	sprintf(fn,"%s.%d",OutName,seedno+1);
	MatSave(result,fn);
	MatFree(result);
	MatFree(seed_vec);
    }

    /* Clean up.
       --------- */
    AppFree(App);
    return 0;
}





/*{{{ --------------------------------------------------------------------------
!section apps.general
!program zsc "Spin-up with script"
!synopsis
    zsb [-QV] [-g <NGen>] <Gen> <Seed> <Op> [<Basis>]
!arg <Gen>
    Generator name.
!arg <Seed>
    Seed space file name.
!arg <Op>
    Spin-up script.
!arg <Basis>
    Output file name, default is {\it Seed}.
!option -g
    Read <NGen> generators, default is 2.
!option -Q
    Quiet, no messages.
!option -V
    Verbose, more messages.
!seealso zsp
!description
This program reads in two or more matrices from `<Gen>.1', `<Gen>.2', ...,
a list of seed vectors from `<Seed>' and a 
list of operations (the scipt), from `<Op>'. Then, ZSC applies the
script to each 
seed vector and writes the output in a separate file (`<Op>.1', 
`<Op>.1', ...) for each vector.

The generators must be square matrices over the same field. `<Seed>' 
must be a matrix over the same field with the same number of columns. 
`<Op>' must be an integer matrix with two columns.

The format of the <Op> file is described together with the ZSP program. 
The <Op> file must be produced with a cyclic seed vector, i.e., the 
first row must be $(1,-1)$ and all other rows must be of the form 
$(m,n)$ with $n != -1$.

The number of generators is 2 by default. This can be changed
by using the `-g' option. The number of output files equals
the number of seed vectors. If no <Out> argument is given,
the output name defaults to <Seed>. For example,
||    zsc -g 3 gen seed op
reads three generators fom `gen.1', `gen.2' and `gen.3',
seed vectors from `seed', the script from `op', and
writes the output to `seed.0001', `seed.0002', ...
-------------------------------------------------------------------------- }}}*/

