/* ============================= C MeatAxe ==================================
   File:        $Id: zzz2.c,v 1.2 2000/12/09 18:26:41 mringe Exp $
   Comment:     Basic row operations.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <string.h>
#include <stdlib.h>

#define LPR (FfCurrentRowSize/sizeof(long))

/* --------------------------------------------------------------------------
   Lokal data
   -------------------------------------------------------------------------- */

MTX_DEFINE_FILE_INFO




/*{{{ --------------------------------------------------------------------------
!section kernel.ff.ff
!variable FfChar  "Characteristic"
!synopsis
    extern int FfChar;
!description
    The |FfChar| variable contains the characteristic of the current field.
    Like |FfOrder|, this variable may be used anywhere, but it must not
    be modified directly.
!seealso  FfOrder FfGen
-------------------------------------------------------------------------- }}}*/

int FfChar = 0;


/*{{{ --------------------------------------------------------------------------
!variable FfOrder  "Field order"
!synopsis
    extern int FfOrder;
!description
    This variable contains the order of the current field. |FfOrder| max be used
    in expressiond but must never modified directly. To change the current 
    field, use |FfSetField()|.
!seealso  FfChar FfGen
-------------------------------------------------------------------------- }}}*/

int FfOrder = -1;


/*{{{ --------------------------------------------------------------------------
!variable FfGen  "Field generator"
!section kernel.ff.ff
!synopsis
    extern FEL FfGen;
!description
    This variable contains a genrator for the multiplicative group of the
    current field.
!seealso  FfChar FfOrder
-------------------------------------------------------------------------- }}}*/

FEL FfGen = 0;



/*{{{ --------------------------------------------------------------------------
!variable FfNoc  "Current row size"
!section kernel.ff.row
!synopsis
    extern int FfNoc;
!description
    This variable contains the current number of columns used for all
    row operations. |FfNoc| is updated automatically when the row
    size is changed with |FfSetNoc()|. 
!seealso FfSetNoc
-------------------------------------------------------------------------- }}}*/

int FfNoc = 0;



/*{{{ --------------------------------------------------------------------------
!function FfAlloc "Allocate memory and initialize"
!synopsis 
    PTR FfAlloc(long nrows);
!arg nrows
    Number of rows.
!return
    Returns a pointer to the memory block or NULL on error.
!description
   This function allocates a block of memory for a vector (if |nrows| is 1)
   or a matrix. Memory is initialized to zero. Field order and row size must 
   have been set with |FfSetField()| and |FfSetNoc()|, respectively. Unlike 
   |malloc()| on some systems, this function does not fail if |nrows| is zero. 
   Instead, a memory block of length zero is allocated.
!seealso FfRowSize FfFree
-------------------------------------------------------------------------- }}}*/

PTR FfAlloc(int nrows)
{
    PTR p;
    char *q;
    register long i;
    size_t req = (size_t)(FfCurrentRowSize*nrows);

    if ((p = (PTR) SysMalloc(req)) == NULL)
    {
	MTX_ERROR3("%E: Cannot allocate %d rows (%d bytes)",MTX_ERR_NOMEM,
	    nrows,(int)req);
	return NULL;
    }

    /* Initialize */
    q = (char *) p;
    for (i = nrows; i > 0; --i)
    {
	FfMulRow((PTR) q,FF_ZERO);
	q += FfCurrentRowSize;
    }
    return p;
}


/*{{{ --------------------------------------------------------------------------
!function FfFree "Free memory"
!arg x
    Pointer to memory.
!description
   This function frees a block of memory that has been allocated with
   |FfAlloc()| before. If the argument is NULL, |FfFree()| does nothing.
!seealso FfAlloc
-------------------------------------------------------------------------- }}}*/

void FfFree(PTR x)
{
    if (x != NULL)
	free(x);
}



/*{{{ --------------------------------------------------------------------------
!function FfCopyRow "Copy a row"
!synopsis 
    void FfCopyRow(PTR dest, PTR src);
!arg dest
    Pointer to the destination.
!arg src
    Pointer to the source.
!description
    This function copies the contents of one row to another row.
    As with all row operations, the row length must have been set before 
    with |FfSetNoc()|.
!seealso FfSwapRows
-------------------------------------------------------------------------- }}}*/

void FfCopyRow(PTR dest, PTR src)

{
    register long *s = (long *) src, *d = (long *) dest, i;
    for (i = LPR; i > 0; --i)
   	 *d++ = *s++;
}



/*{{{ --------------------------------------------------------------------------
!function FfSwapRows "Exchange two rows"
!synopsis 
    void FfSwapRows(PTR dest, PTR src);
!arg dest
    Pointer to the first row
!arg src
    Pointer to the second row
!description
    This function exchanges the contents of two rows. As with all row 
    operations, the row length must have been set before with |FfSetNoc()|.
!seealso FfCopyRow
-------------------------------------------------------------------------- }}}*/

void FfSwapRows(PTR dest, PTR src)

{	register long *p1 = (long *) src, *p2 = (long *) dest, i, l;

	for (i = LPR; i > 0; --i)
	{	l = *p1;
		*p1++ = *p2;
		*p2++ = l;
	}
}



/*{{{ --------------------------------------------------------------------------
!function FfGetPtr "Get pointer to a row"
!arg base
    Pointer to the first row of the matrix.
!arg row
    Row index.
!description
    This function returns a pointer to the |nor|-th row of a matrix, assuming
    the current row size. |base| must be a pointer to the beginning of a row, 
    but this need not be the first row of the matrix. For example, the 
    following code steps through the odd rows of a matrix:
    ||    PTR r = matrix;
    ||    int i;
    ||    for (i = 1; i < nrows; i += 2)
    ||    {
    ||        r = FfGetPtr(r,2);
    ||        ...
    ||    }
    Note: The function does not check if the resulting pointer
    is still inside the matrix. 
!seealso FfStepPtr
-------------------------------------------------------------------------- }}}*/

PTR FfGetPtr(PTR base, int row)

{
#ifdef PARANOID
    if (row < 0)
	MTX_ERROR1("row=%d < 0",row);
#endif
    return (PTR) ((char *)base + FfCurrentRowSize * row);
}


/*{{{ --------------------------------------------------------------------------
!function FfStepPtr "Step a row pointer"
!arg x
    Pointer to the |PTR| to increment.
!return
    Always 0.
!description
    This function increments a pointer by 1 row. The row size must have been
    set before with |FfSetNoc()|. |FfStepPtr(&x)| is equivalent to
    |x = FfGetPtr(x,1);|. It is typically used to step through the rows of 
    a matrix. Here is an example with a $100\times 40$ matrix:
    ||    PTR r, mat;
    ||    int i;
    ||    FfSetNoc(40);
    ||    mat = FfAlloc(100);
    ||    for (i = 1, r = mat; i < 100; ++i, FfStepPtr(&r))
    ||        ...
!seealso FfGetPtr
-------------------------------------------------------------------------- }}}*/

int FfStepPtr(PTR *x)

{
    *(char **)x += FfCurrentRowSize;
    return 0;
}




